##
##

import attr
import logging
import socket
import time
import os
import jinja2
import json
from typing import Optional, List
from couchformation.exception import NonFatalError
from couchformation.config import NodeList, get_state_dir
from couchformation.provisioner.ssh import RunSSHCommand
from couchformation.provisioner.sftp import SFTPFile, SFTPFileObject
import couchformation.constants as C

logger = logging.getLogger('couchformation.provisioner.remote')
logger.addHandler(logging.NullHandler())
logging.getLogger("paramiko").setLevel(logging.ERROR)


class ProvisionerError(NonFatalError):
    pass


class CustomLogFormatter(logging.Formatter):
    FORMATS = {
        logging.DEBUG: f"{C.FORMAT_TIMESTAMP} [{C.FORMAT_LEVEL}] {C.FORMAT_MESSAGE}",
        logging.INFO: f"{C.FORMAT_TIMESTAMP} [{C.FORMAT_LEVEL}] {C.FORMAT_MESSAGE}",
        logging.WARNING: f"{C.FORMAT_TIMESTAMP} [{C.FORMAT_LEVEL}] {C.FORMAT_MESSAGE}",
        logging.ERROR: f"{C.FORMAT_TIMESTAMP} [{C.FORMAT_LEVEL}] {C.FORMAT_MESSAGE}",
        logging.CRITICAL: f"{C.FORMAT_TIMESTAMP} [{C.FORMAT_LEVEL}] {C.FORMAT_MESSAGE}"
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        if logging.DEBUG >= logging.root.level:
            log_fmt += C.FORMAT_EXTRA
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


@attr.s
class ProvisionSet:
    pre_install_cmd: Optional[List[str]] = attr.ib(default=[])
    install_cmd: Optional[List[str]] = attr.ib(default=[])
    post_install_cmd: Optional[List[str]] = attr.ib(default=[])
    nodes: Optional[NodeList] = attr.ib(default=None)

    @classmethod
    def create(cls):
        return cls(
            [],
            [],
            [],
            None
        )

    def add_pre_install(self, pre_install_cmds: List[str]):
        self.pre_install_cmd.extend(pre_install_cmds)

    def add_install(self, install_cmds: List[str]):
        self.install_cmd.extend(install_cmds)

    def add_post_install(self, post_install_cmds: List[str]):
        self.post_install_cmd.extend(post_install_cmds)

    def add_nodes(self, node_list: NodeList):
        self.nodes = node_list


class RemoteProvisioner(object):

    def __init__(self, parameters: dict, command: str = '', root: bool = True):
        self.parameters = parameters
        self.command = command
        self.root = root
        self.service = self.parameters.get('service')
        self.project = self.parameters.get('project')
        self.public_ip = self.parameters.get('public_ip')
        self.private_ip = self.parameters.get('private_ip')
        self.public_hostname = self.parameters.get('public_hostname')
        self.private_hostname = self.parameters.get('private_hostname')
        self.public = self.parameters.get('public') if 'public' in self.parameters else False
        self.username = self.parameters.get('username')
        self.ssh_key = self.parameters.get('ssh_key')
        self.zone = self.parameters.get('zone')
        self.password = self.parameters.get('password') if 'password' in self.parameters else 'password'
        self.private_key = self.parameters.get('private_key') if 'private_key' in self.parameters else 'null'
        self.host_password = self.parameters.get('host_password')
        self.upload_file = self.parameters.get('upload')
        self.sw_version = self.parameters.get('sw_version') if 'sw_version' in self.parameters else 'latest'
        self.services = self.parameters.get('services')
        self.options = self.parameters.get('options') if self.parameters.get('options') else 'default'
        self.connect = ','.join(self.parameters.get('connect')) \
            if self.parameters.get('connect') and type(self.parameters.get('connect')) is list \
            else self.parameters.get('connect')
        self.private_ip_list = ','.join(self.parameters.get('private_ip_list'))
        self.public_ip_list = ','.join(self.parameters.get('public_ip_list'))
        self.service_list = ':'.join(self.parameters.get('service_list'))
        if self.parameters.get('private_host_list') and len(self.parameters.get('private_host_list')) > 0:
            self.private_host_list = ','.join(self.parameters.get('private_host_list'))
        else:
            self.private_host_list = 'null'
        if self.parameters.get('public_host_list') and len(self.parameters.get('public_host_list')) > 0:
            self.public_host_list = ','.join(self.parameters.get('public_host_list'))
        else:
            self.public_host_list = 'null'
        self.use_private_ip = self.parameters.get('use_private_ip') if self.parameters.get('use_private_ip') else False

        logger.debug(f"Parameters:\n{json.dumps(self.parameters, indent=2)}")

    def upload(self):
        if self.use_private_ip:
            hostname = self.private_ip
        else:
            hostname = self.public_ip

        filename = os.path.basename(self.upload_file)

        if not self.wait_port(hostname):
            raise ProvisionerError(f"Host {hostname} is not reachable")

        logger.info(f"Connection to {hostname} successful")
        time.sleep(0.5)

        SFTPFile(self.ssh_key, self.username, hostname, self.upload_file, f"/var/tmp/{filename}").upload()

        return 0

    def copy_file(self, fl, target):
        if self.use_private_ip:
            hostname = self.private_ip
        else:
            hostname = self.public_ip

        if not self.wait_port(hostname):
            raise ProvisionerError(f"Host {hostname} is not reachable")

        logger.info(f"Connection to {hostname} successful")
        time.sleep(0.5)

        SFTPFileObject(self.ssh_key, self.username, hostname, fl, target).copy_file()

        return 0

    def run(self):
        working_dir = get_state_dir(self.project, self.service)
        file_output = logging.getLogger('couchformation.provisioner.output')
        file_output.propagate = False
        log_file = os.path.join(working_dir, 'provision.log')
        file_handler = logging.FileHandler(log_file)
        file_handler.setFormatter(CustomLogFormatter())
        file_output.addHandler(file_handler)
        file_output.setLevel(logging.DEBUG)

        if self.use_private_ip:
            hostname = self.private_ip
        else:
            hostname = self.public_ip

        if self.root:
            command = f"""sudo {self.command}"""
        else:
            command = self.command

        if not self.wait_port(hostname):
            raise ProvisionerError(f"Host {hostname} is not reachable")

        logger.info(f"Connection to {hostname} successful")
        time.sleep(0.5)

        _command = self.resolve_variables(command)

        logger.info(f"Connecting to {hostname} as {self.username}")
        logger.debug(f"Using SSH key {self.ssh_key}")
        logger.debug(f"Running command: {_command}")

        file_output.info(f"{hostname}: [{_command}] begins")

        exit_code, stdout, stderr = RunSSHCommand(self.ssh_key, self.username, hostname, _command, working_dir).exec()

        for line in stdout.readlines():
            line_out = line.strip()
            log_out = f"{hostname}: {line_out}"
            logger.info(log_out)
            file_output.info(log_out)

        file_output.info(f"{hostname}: [{_command}] complete")

        logger.info(f"Command complete for host {hostname}")
        logger.debug(f"[{_command}] returned {exit_code} on {hostname}")

        file_output.removeHandler(file_handler)
        file_handler.close()
        return exit_code

    @staticmethod
    def wait_port(address: str, port: int = 22, retry_count=300, factor=0.1):
        for retry_number in range(retry_count + 1):
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(1)
            result = sock.connect_ex((address, port))
            sock.close()
            if result == 0:
                return True
            else:
                if retry_number == retry_count:
                    return False
                logger.info(f"Waiting for {address} to become reachable")
                wait = factor
                wait *= (2 ** (retry_number + 1))
                time.sleep(wait)

    def resolve_variables(self, line: str):
        env = jinja2.Environment(undefined=jinja2.DebugUndefined)
        raw_template = env.from_string(line)
        formatted_value = raw_template.render(
            SERVICE_NAME=self.service,
            SOFTWARE_VERSION=self.sw_version,
            PASSWORD=self.password,
            PRIVATE_KEY=self.private_key,
            PRIVATE_IP_LIST=self.private_ip_list,
            PUBLIC_IP_LIST=self.public_ip_list,
            SERVICE_LIST=self.service_list,
            OPTIONS=self.options,
            IP_LIST=self.public_ip_list if self.public else self.private_ip_list,
            HOST_LIST=self.public_host_list if self.public else self.private_host_list,
            NODE_ZONE=self.zone,
            SERVICES=self.services,
            CONNECT_SERVICE=self.connect,
            CONNECT_IP=self.connect,
            CONNECT_LIST=self.connect if self.connect else '127.0.0.1'
        )
        return formatted_value



##
##

import logging
import botocore.exceptions
from couchformation.aws.driver.base import CloudBase, AWSDriverError

logger = logging.getLogger('couchformation.aws.driver.dns')
logger.addHandler(logging.NullHandler())
logging.getLogger("botocore").setLevel(logging.ERROR)
logging.getLogger("urllib3").setLevel(logging.ERROR)


class DNS(CloudBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def create(self, domain: str, vpc_id: str = None, region: str = None):
        private = vpc_id is not None
        zone_region = region if region is not None else self.aws_region
        kwargs = {
            'HostedZoneConfig': {
                'PrivateZone': private
            },
            'CallerReference': str(hash(domain + '_private_' + str(private)))
        }
        if private:
            kwargs['VPC'] = {
                'VPCRegion': zone_region,
                'VPCId': vpc_id
            }
        try:
            result = self.dns_client.create_hosted_zone(Name=domain, **kwargs)
            return result.get('HostedZone', {}).get('Id')
        except Exception as err:
            raise AWSDriverError(f"error creating hosted domain: {err}")

    def details(self, hosted_zone: str):
        try:
            result = self.dns_client.get_hosted_zone(Id=hosted_zone)
            return result.get('HostedZone', {})
        except botocore.exceptions.ClientError as err:
            if err.response['Error']['Code'] == 'NoSuchHostedZone':
                return None
            raise AWSDriverError(f"ClientError: {err}")
        except Exception as err:
            raise AWSDriverError(f"error: {err}")

    def zone_id(self, domain: str):
        try:
            result = self.dns_client.list_hosted_zones()
            r_set = next((item for item in result.get('HostedZones', [])
                          if item.get('Name').startswith(domain) and item.get('Config', {}).get('PrivateZone', False) is False), None)
            return r_set.get('Id') if r_set else None
        except botocore.exceptions.ClientError as err:
            if err.response['Error']['Code'] == 'NoSuchHostedZone':
                return None
            raise AWSDriverError(f"ClientError: {err}")
        except Exception as err:
            raise AWSDriverError(f"error: {err}")

    def record_sets(self, hosted_zone: str, r_type: str):
        try:
            result = self.dns_client.list_resource_record_sets(HostedZoneId=hosted_zone)
            r_set = next((item for item in result.get('ResourceRecordSets', {}) if item.get('Type') == r_type), None)
            return [item.get('Value') for item in r_set.get('ResourceRecords', [])]
        except botocore.exceptions.ClientError as err:
            if err.response['Error']['Code'] == 'NoSuchHostedZone':
                return None
            raise AWSDriverError(f"ClientError: {err}")
        except Exception as err:
            raise AWSDriverError(f"error: {err}")

    def delete(self, hosted_zone: str):
        try:
            result = self.dns_client.delete_hosted_zone(Id=hosted_zone)
            return result.get('ChangeInfo', {}).get('Status')
        except Exception as err:
            raise AWSDriverError(f"error deleting hosted domain: {err}")

    def add_record(self, hosted_zone: str, name: str, values: list, record_type: str = 'A', ttl: int = 300):
        change_batch = {
            'Changes': [
                {
                    'Action': 'CREATE',
                    'ResourceRecordSet': {
                        'Name': name,
                        'Type': record_type,
                        'TTL': ttl,
                        'ResourceRecords': []
                    }
                }
            ]
        }
        for item in values:
            change_batch['Changes'][0]['ResourceRecordSet']['ResourceRecords'].append({'Value': item})
        try:
            result = self.dns_client.change_resource_record_sets(HostedZoneId=hosted_zone, ChangeBatch=change_batch)
            return result.get('ChangeInfo', {}).get('Status')
        except Exception as err:
            raise AWSDriverError(f"error adding record to domain: {err}")

    def delete_record(self, hosted_zone: str, name: str, values: list, record_type: str = 'A', ttl: int = 300):
        change_batch = {
            'Changes': [
                {
                    'Action': 'DELETE',
                    'ResourceRecordSet': {
                        'Name': name,
                        'Type': record_type,
                        'TTL': ttl,
                        'ResourceRecords': []
                    }
                }
            ]
        }
        for item in values:
            change_batch['Changes'][0]['ResourceRecordSet']['ResourceRecords'].append({'Value': item})
        try:
            result = self.dns_client.change_resource_record_sets(HostedZoneId=hosted_zone, ChangeBatch=change_batch)
            return result.get('ChangeInfo', {}).get('Status')
        except Exception as err:
            raise AWSDriverError(f"error deleting record from domain: {err}")

    def list_associations(self, vpc_id: str, region: str):
        extra_args = {}
        results = []
        try:
            while True:
                result = self.dns_client.list_hosted_zones_by_vpc(VPCId=vpc_id, VPCRegion=region, **extra_args)
                results.extend(result['HostedZoneSummaries'])
                if 'NextToken' not in result:
                    break
                extra_args['NextToken'] = result['NextToken']
            return results
        except Exception as err:
            raise AWSDriverError(f"error getting VPC list: {err}")

    def associate(self, hosted_zone: str, vpc_id: str, region: str):
        current_associations = self.list_associations(vpc_id, region)
        if next((z for z in current_associations if z.get('HostedZoneId') == hosted_zone), None):
            return
        vpc_info = {
            'VPCRegion': region,
            'VPCId': vpc_id
        }
        try:
            result = self.dns_client.associate_vpc_with_hosted_zone(HostedZoneId=hosted_zone, VPC=vpc_info)
            return result.get('ChangeInfo', {}).get('Status')
        except Exception as err:
            raise AWSDriverError(f"error associating hosted zone: {err}")

    def disassociate(self, hosted_zone: str, vpc_id: str, region: str):
        current_associations = self.list_associations(vpc_id, region)
        if not next((z for z in current_associations if z.get('HostedZoneId') == hosted_zone), None):
            return
        vpc_info = {
            'VPCRegion': region,
            'VPCId': vpc_id
        }
        try:
            result = self.dns_client.disassociate_vpc_from_hosted_zone(HostedZoneId=hosted_zone, VPC=vpc_info)
            return result.get('ChangeInfo', {}).get('Status')
        except botocore.exceptions.ClientError as err:
            if err.response['Error']['Code'] == 'NoSuchHostedZone':
                return
            raise AWSDriverError(f"ClientError: {err}")
        except Exception as err:
            raise AWSDriverError(f"error associating hosted zone: {err}")



##
##

import os
import logging
import random
import string
from itertools import cycle
from typing import List
from couchformation.network import NetworkDriver
from couchformation.azure.driver.network import Network, Subnet, SecurityGroup
from couchformation.azure.driver.base import CloudBase
from couchformation.azure.driver.dns import DNS
from couchformation.azure.driver.private_dns import PrivateDNS
import couchformation.azure.driver.constants as C
from couchformation.config import get_state_file, get_state_dir, PortSettingSet, PortSettings, State
from couchformation.deployment import MetadataManager
from couchformation.exception import FatalError
from couchformation.kvdb import KeyValueStore
from couchformation.util import FileManager, synchronize, dump_class_variables


logger = logging.getLogger('couchformation.azure.network')
logger.addHandler(logging.NullHandler())


class AzureNetworkError(FatalError):
    pass


class AzureNetwork(object):

    def __init__(self, parameters: dict):
        self.parameters = parameters
        self.name = parameters.get('name')
        self.project = parameters.get('project')
        self.region = parameters.get('region')
        self.auth_mode = parameters.get('auth_mode')
        self.profile = parameters.get('profile')
        self.ssh_key = parameters.get('ssh_key')
        self.cloud = parameters.get('cloud')
        self.domain = parameters.get('domain')
        self.allow = parameters.get('allow') if parameters.get('allow') else "0.0.0.0/0"
        self.build_ports = PortSettingSet().create().items()

        filename = get_state_file(self.project, f"network-{self.region}")

        try:
            state_dir = get_state_dir(self.project, f"network-{self.region}")
            if not os.path.exists(state_dir):
                FileManager().make_dir(state_dir)
        except Exception as err:
            raise AzureNetworkError(f"can not create state dir: {err}")

        document = f"network:{self.cloud}"
        self.state = KeyValueStore(filename, document)

        self.az_network = Network(self.parameters)
        self.az_base = CloudBase(self.parameters)

        project_uid = MetadataManager(self.project).project_uid
        self.asset_prefix = f"cf-{project_uid}"
        self.rg_name = f"{self.asset_prefix}-rg"
        self.vpc_name = f"{self.asset_prefix}-vpc"
        self.nsg_name = f"{self.asset_prefix}-nsg"
        self.subnet_name = f"{self.asset_prefix}-subnet-01"
        self.vnet_dns_link_name = f"{self.asset_prefix}-dns-link"

    def check_state(self):
        if self.state.get('resource_group'):
            rg_name = self.state.get('resource_group')
        else:
            rg_name = self.rg_name

        if self.state.get('network'):
            vpc_name = self.state.get('network')
        else:
            vpc_name = self.vpc_name

        if self.state.get('subnet'):
            result = Subnet(self.parameters).details(vpc_name, self.state['subnet'], rg_name)
            if result is None:
                logger.warning(f"Removing stale state entry for subnet {self.state['subnet']}")
                del self.state['subnet']
                del self.state['subnet_id']
                del self.state['subnet_cidr']
        else:
            result = Subnet(self.parameters).details(vpc_name, self.subnet_name, rg_name)
            if result:
                logger.warning(f"Importing orphaned entry for subnet {self.subnet_name}")
                self.state['subnet'] = self.subnet_name
                self.state['subnet_id'] = result['id']
                self.state['subnet_cidr'] = result['cidr']

        if self.state.get('network_security_group'):
            result = SecurityGroup(self.parameters).details(self.state['network_security_group'], rg_name)
            if result is None:
                logger.warning(f"Removing stale state entry for security group {self.state['network_security_group']}")
                del self.state['network_security_group']
                del self.state['network_security_group_id']
        else:
            result = SecurityGroup(self.parameters).details(self.nsg_name, rg_name)
            if result:
                logger.warning(f"Importing orphaned entry for network security group {self.nsg_name}")
                self.state['network_security_group'] = self.nsg_name
                self.state['network_security_group_id'] = result['id']

        for sg_rule_key in self.state.key_match('rule_.*'):
            result = SecurityGroup(self.parameters).search_rules(self.state['network_security_group'], rg_name, self.state.get(sg_rule_key))
            if result is None:
                logger.warning(f"Removing stale state entry for security group rule {self.state[sg_rule_key]}")
                del self.state[sg_rule_key]

        if self.state.get('network'):
            result = Network(self.parameters).details(self.state['network'], rg_name)
            if result is None:
                logger.warning(f"Removing stale state entry for network {self.state['network']}")
                del self.state['network']
                del self.state['network_cidr']
                del self.state['network_id']
        else:
            result = Network(self.parameters).details(self.vpc_name, rg_name)
            if result:
                logger.warning(f"Importing orphaned entry for network {self.vpc_name}")
                self.state['network'] = self.vpc_name
                self.state['network_id'] = result.id
                self.state['network_cidr'] = result.address_space.address_prefixes[0]

        if self.state.get('public_hosted_zone'):
            result = DNS(self.parameters).details(self.state['public_hosted_zone'])
            if result is None:
                logger.warning(f"Removing stale state entry for public managed zone {self.state['public_hosted_zone']}")
                del self.state['public_hosted_zone']

        if self.state.get('private_dns_zone_link') and self.state['private_hosted_zone']:
            result = PrivateDNS(self.parameters).vpc_link_details(self.state['private_hosted_zone'], self.state['private_dns_zone_link'], rg_name)
            if result is None:
                logger.warning(f"Removing stale state entry for private DNS zone link {self.state['private_dns_zone_link']}")
                del self.state['private_dns_zone_link']

        if self.state.get('private_hosted_zone'):
            result = PrivateDNS(self.parameters).details(self.state['private_hosted_zone'])
            if result is None:
                logger.warning(f"Removing stale state entry for private managed zone {self.state['private_hosted_zone']}")
                del self.state['private_hosted_zone']

        if self.state.get('resource_group'):
            result = self.az_base.get_rg(self.state['resource_group'], self.az_base.region)
            if result is None:
                logger.warning(f"Removing stale state entry for resource group {self.state['resource_group']}")
                del self.state['resource_group']
                del self.state['zone']
        else:
            result = self.az_base.get_rg(self.rg_name, self.az_base.region)
            if result:
                logger.warning(f"Importing orphaned entry for resource group {self.rg_name}")
                self.state['resource_group'] = self.rg_name

    @synchronize()
    def create_vpc(self):
        self.check_state()
        cidr_util = NetworkDriver()

        for net in self.az_network.cidr_list:
            cidr_util.add_network(net)

        zone_list = self.az_network.zones()
        azure_location = self.az_base.region

        try:
            self.state['state'] = State.DEPLOYING.value
            logger.debug(f"Azure VNet create input variables:\n{dump_class_variables(vars(self))}")

            if not self.state.get('resource_group'):
                self.az_base.create_rg(self.rg_name, azure_location)
                self.state['resource_group'] = self.rg_name
                logger.info(f"Created resource group {self.rg_name}")
            else:
                self.rg_name = self.state['resource_group']

            if not self.state.get('network'):
                vpc_cidr = cidr_util.get_next_network()
                net_resource = Network(self.parameters).create(self.vpc_name, vpc_cidr, self.rg_name)
                net_resource_id = net_resource.id
                self.state['network'] = self.vpc_name
                self.state['network_cidr'] = vpc_cidr
                self.state['network_id'] = net_resource_id
                logger.info(f"Created network {self.vpc_name}")
            else:
                self.vpc_name = self.state['network']
                vpc_cidr = self.state['network_cidr']
                net_resource_id = self.state['network_id']
                cidr_util.set_active_network(vpc_cidr)

            if not self.state.get('network_security_group'):
                nsg_resource = SecurityGroup(self.parameters).create(self.nsg_name, self.rg_name)
                nsg_resource_id = nsg_resource.id
                SecurityGroup(self.parameters).add_rule("AllowSSH", self.nsg_name, ["22"], 100, self.rg_name)
                self.state['network_security_group'] = self.nsg_name
                self.state['network_security_group_id'] = nsg_resource_id
                logger.info(f"Created network security group {self.nsg_name}")
            else:
                nsg_resource_id = self.state['network_security_group_id']

            subnet_list = list(cidr_util.get_next_subnet())
            subnet_cycle = cycle(subnet_list[1:])

            if not self.state.get('subnet_cidr'):
                subnet_cidr = next(subnet_cycle)
                self.state['subnet_cidr'] = subnet_cidr
            else:
                subnet_cidr = self.state['subnet_cidr']

            if not self.state.get('subnet'):
                subnet_resource = Subnet(self.parameters).create(self.subnet_name, self.vpc_name, subnet_cidr, nsg_resource_id, self.rg_name)
                subnet_id = subnet_resource.id
                self.state['subnet'] = self.subnet_name
                self.state['subnet_id'] = subnet_id
                logger.info(f"Created subnet {self.subnet_name}")
            else:
                subnet_id = self.state['subnet_id']
                self.subnet_name = self.state['subnet']

            for n, zone in enumerate(zone_list):
                if self.state.list_exists('zone', zone):
                    continue
                self.state.list_add('zone', zone, self.subnet_name, subnet_id)
                logger.info(f"Added zone {zone}")

            if self.domain and not self.state.get('domain'):
                domain_prefix = ''.join(random.choice(string.ascii_lowercase) for _ in range(7))
                domain_name = f"{domain_prefix}.{self.domain}"
                self.state['domain'] = domain_name
                logger.info(f"Generated project domain {domain_name}")
            elif self.state.get('domain'):
                domain_name = self.state.get('domain')
                logger.info(f"Using existing domain {domain_name}")
            else:
                domain_name = None

            if domain_name and not self.state.get('public_hosted_zone'):
                domain_id = DNS(self.parameters).create(domain_name, self.rg_name)
                self.state['public_hosted_zone'] = domain_id
                logger.info(f"Created public managed zone {domain_id} for domain {domain_name}")

            if domain_name and not self.state.get('private_hosted_zone'):
                domain_id = PrivateDNS(self.parameters).create(domain_name, self.rg_name)
                self.state['private_hosted_zone'] = domain_id
                logger.info(f"Created private managed zone {domain_id} for domain {domain_name}")

            if self.state.get('private_hosted_zone') and not self.state.get('private_dns_zone_link'):
                PrivateDNS(self.parameters).vpc_link(domain_name, self.vnet_dns_link_name, net_resource_id, self.rg_name)
                self.state['private_dns_zone_link'] = self.vnet_dns_link_name
                logger.info(f"Created private DNS zone link {self.vnet_dns_link_name}")

            if self.state.get('public_hosted_zone') and not self.state['parent_hosted_zone']:
                parent_domain = '.'.join(domain_name.split('.')[1:])
                parent_id = DNS(self.parameters).zone_name(parent_domain)
                parent_rg = DNS(self.parameters).zone_rg(parent_domain)
                if parent_id:
                    ns_names = DNS(self.parameters).record_sets(self.state['public_hosted_zone'], 'NS', self.rg_name)
                    DNS(self.parameters).add_record(parent_id, domain_name, ns_names, parent_rg, 'NS')
                    self.state['parent_hosted_zone'] = parent_id
                    self.state['parent_hosted_zone_rg'] = parent_rg
                    self.state['parent_zone_ns_records'] = ','.join(ns_names)
                    logger.info(f"Added {len(ns_names)} NS record(s) to domain {parent_domain}")

            self.state['state'] = State.DEPLOYED.value
        except Exception as err:
            raise AzureNetworkError(f"Error creating network: {err}")

    @synchronize()
    def create_build_sg(self, build_name: str):
        nsg_name = self.network_security_group
        rg_name = self.resource_group
        for build_port_cfg in self.build_ports:
            if build_port_cfg.build != build_name:
                continue
            state_key_name = f"rule_{build_name}"
            build_rule_name = f"Allow{build_name.upper()}"
            if not self.state.get(state_key_name):
                if build_port_cfg.has_tcp_ports:
                    SecurityGroup(self.parameters).add_rule(f"{build_rule_name}_TCP", nsg_name, list(build_port_cfg.tcp_as_ranges()), 0, rg_name, "tcp")
                if build_port_cfg.has_udp_ports:
                    SecurityGroup(self.parameters).add_rule(f"{build_rule_name}_UDP", nsg_name, list(build_port_cfg.udp_as_ranges()), 0, rg_name, "udp")
                self.state[state_key_name] = build_rule_name
                logger.info(f"Added NSG rule {build_rule_name}")

    @synchronize()
    def create_win_sg(self):
        nsg_name = self.network_security_group
        rg_name = self.resource_group
        if not self.state.get('rule_win_rdp'):
            rule_name = "AllowRDP"
            SecurityGroup(self.parameters).add_rule(rule_name, nsg_name, ["3389", "5985", "5986"], 0, rg_name, "tcp")
            self.state['rule_win_rdp'] = rule_name
            logger.info(f"Added NSG rule {rule_name}")

    @synchronize()
    def create_node_group_sg(self, service: str, group: int, ports: List[str]):
        nsg_name = self.network_security_group
        rg_name = self.resource_group
        state_key_name = f"rule_{service}_group_{group}"
        rule_name = f"Allow{service.upper()}{group:02d}"
        if not self.state.get(state_key_name):
            port_cfg = PortSettings().create(self.name, ports)
            if port_cfg.has_tcp_ports:
                SecurityGroup(self.parameters).add_rule(f"{rule_name}_TCP", nsg_name, list(port_cfg.tcp_as_ranges()), 0, rg_name, "tcp")
            if port_cfg.has_udp_ports:
                SecurityGroup(self.parameters).add_rule(f"{rule_name}_UDP", nsg_name, list(port_cfg.udp_as_ranges()), 0, rg_name, "udp")
            self.state[state_key_name] = rule_name
            logger.info(f"Added NSG rule {rule_name}")

    @synchronize()
    def peer_vpc(self):
        logger.warning(f"Peering not implemented for cloud {self.cloud}")

    @synchronize()
    def unpeer_vpc(self):
        pass

    @synchronize()
    def destroy_vpc(self):
        if self.state.list_len('services') > 0:
            logger.info(f"Active services, leaving project network in place")
            return

        self.check_state()
        try:
            self.state['state'] = State.DESTROYING.value

            if self.state.get('resource_group'):
                rg_name = self.state.get('resource_group')
            else:
                logger.warning("No saved resource group")
                rg_name = self.rg_name

            if self.state.get('network'):
                vpc_name = self.state.get('network')
            else:
                logger.warning("No saved network")
                return

            if self.state.get('subnet'):
                subnet_name = self.state.get('subnet')
                Subnet(self.parameters).delete(vpc_name, subnet_name, rg_name)
                del self.state['subnet']
                del self.state['subnet_id']
                logger.info(f"Removed subnet {subnet_name}")

            if self.state.get('subnet_cidr'):
                del self.state['subnet_cidr']

            if self.state.get('network_security_group'):
                nsg_name = self.state.get('network_security_group')
                SecurityGroup(self.parameters).delete(nsg_name, rg_name)
                del self.state['network_security_group']
                del self.state['network_security_group_id']
                logger.info(f"Removed network security group {nsg_name}")

            for sg_rule_key in self.state.key_match('rule_.*'):
                rule_name = self.state.get(sg_rule_key)
                del self.state[sg_rule_key]
                logger.info(f"Removed security group rule {rule_name}")

            for n, zone_state in reversed(list(enumerate(self.state.list_get('zone')))):
                self.state.list_remove('zone', zone_state[0])

            if self.state.get('parent_hosted_zone') and self.state.get('domain'):
                DNS(self.parameters).delete_record(self.state['parent_hosted_zone'], self.state['domain'], self.state['parent_hosted_zone_rg'], 'NS')
                del self.state['parent_hosted_zone']
                del self.state['parent_zone_ns_records']
                logger.info(f"Removing NS records for domain {self.state['domain']}")

            if self.state.get('public_hosted_zone'):
                domain_id = self.state.get('public_hosted_zone')
                DNS(self.parameters).delete(domain_id, rg_name)
                del self.state['public_hosted_zone']
                logger.info(f"Removing public hosted zone {domain_id}")

            if self.state.get('private_dns_zone_link'):
                domain_id = self.state.get('private_hosted_zone')
                PrivateDNS(self.parameters).vpc_unlink(domain_id, self.state['private_dns_zone_link'], rg_name)
                del self.state['private_dns_zone_link']
                logger.info(f"Removing private DNS zone link for {domain_id}")

            if self.state.get('private_hosted_zone'):
                domain_id = self.state.get('private_hosted_zone')
                PrivateDNS(self.parameters).delete(domain_id, rg_name)
                del self.state['private_hosted_zone']
                logger.info(f"Removing private hosted zone {domain_id}")

            if self.state.get('network'):
                vpc_name = self.state.get('network')
                Network(self.parameters).delete(vpc_name, rg_name)
                del self.state['network']
                del self.state['network_cidr']
                del self.state['network_id']
                logger.info(f"Removed network {vpc_name}")

            if self.state.get('domain'):
                domain_name = self.state.get('domain')
                del self.state['domain']
                logger.info(f"Removing project domain {domain_name}")

            if self.state.get('resource_group'):
                rg_name = self.state.get('resource_group')
                self.az_base.delete_rg(rg_name)
                del self.state['resource_group']
                logger.info(f"Removed resource group {rg_name}")

            self.state['state'] = State.IDLE.value
        except Exception as err:
            raise AzureNetworkError(f"Error removing network: {err}")

    def create(self):
        logger.info(f"Creating cloud network for {self.project} in {C.CLOUD_KEY.upper()}")
        self.create_vpc()

    def destroy(self):
        logger.info(f"Removing cloud network for {self.project} in {C.CLOUD_KEY.upper()}")
        self.destroy_vpc()

    def get(self, key):
        return self.state.get(key)

    @property
    def network(self):
        return self.state.get('network')

    @property
    def subnet(self):
        return self.state.get('subnet')

    @property
    def network_security_group(self):
        return self.state.get('network_security_group')

    @property
    def resource_group(self):
        return self.state.get('resource_group')

    @property
    def zones(self):
        return self.state.list_get('zone')

    @property
    def domain_name(self):
        return self.state.get('domain')

    @property
    def public_zone(self):
        return self.state.get('public_hosted_zone')

    @property
    def private_zone(self):
        return self.state.get('private_hosted_zone')

    def add_service(self, name):
        self.state.list_add('services', name)

    def remove_service(self, name):
        self.state.list_remove('services', name)



##
##

import logging
from couchformation.docker.driver.container import Container
from couchformation.config import get_state_file, get_state_dir, PortSettingSet, PortSettings, State
from couchformation.docker.network import DockerNetwork
from couchformation.exception import FatalError
from couchformation.kvdb import KeyValueStore
from couchformation.util import FileManager, Synchronize
from couchformation.network import NetworkUtil

logger = logging.getLogger('couchformation.docker.node')
logger.addHandler(logging.NullHandler())
logging.getLogger("docker").setLevel(logging.WARNING)


class DockerNodeError(FatalError):
    pass


class DockerDeployment(object):

    def __init__(self, parameters: dict):
        self.parameters = parameters
        self.name = parameters.get('name')
        self.project = parameters.get('project')
        self.build = parameters.get('build')
        self.image = self.build if Container(self.parameters).map(self.build) else parameters.get('image')
        self.number = parameters.get('number')
        self.services = parameters.get('services') if parameters.get('services') else "default"
        self.node_name = f"{self.name}-node-{self.number:02d}"

        filename = get_state_file(self.project, self.name)

        with Synchronize():
            try:
                state_dir = get_state_dir(self.project, self.name)
                FileManager().make_dir(state_dir)
            except Exception as err:
                raise DockerNodeError(f"can not create state dir: {err}")

        document = self.node_name
        self.state = KeyValueStore(filename, document)

        self.docker_network = DockerNetwork(self.parameters)

    def check_state(self):
        if self.state.get('instance_id'):
            result = Container(self.parameters).get_container(self.state.get('instance_id'))
            if result is None:
                logger.warning(f"Removing stale state entry for container {self.state.get('instance_id')}")
                del self.state['instance_id']

    def deploy(self):
        self.check_state()
        if self.state.get('instance_id'):
            logger.info(f"Node {self.node_name} already exists")
            return self.state.as_dict

        self.state['state'] = State.DEPLOYING.value

        net_name = self.docker_network.network
        services = self.services

        build_ports = PortSettingSet().create().get(self.build)
        if build_ports:
            port_cfg = PortSettings().create(self.name, build_ports.tcp_ports)
            ports = ','.join(list(port_cfg.tcp_as_ranges()))
        else:
            ports = None

        logger.info(f"Creating container {self.node_name}")
        container = Container(self.parameters).run(self.image, self.node_name, network=net_name, ports=ports)

        public_ip = NetworkUtil().local_ip_address()
        private_ip = Container(self.parameters).get_container_ip(self.node_name)
        self.state['service'] = self.name
        self.state['instance_id'] = container.id
        self.state['name'] = self.node_name
        self.state['services'] = services
        self.state['public_ip'] = public_ip if public_ip else private_ip
        self.state['private_ip'] = private_ip
        self.docker_network.add_service(self.node_name)

        logger.info(f"Created container {self.node_name}")
        self.state['state'] = State.DEPLOYED.value
        return self.state.as_dict

    def destroy(self):
        self.state['state'] = State.DESTROYING.value
        if self.state.get('instance_id'):
            Container(self.parameters).terminate(self.node_name)
            self.state.clear()
            self.docker_network.remove_service(self.node_name)
            logger.info(f"Removed container {self.node_name}")
        self.state['state'] = State.IDLE.value

    def info(self):
        return self.state.as_dict



##
##

import attr
import os
from pathlib import Path

CLOUD_KEY = "gcp"


def get_auth_directory():
    if os.name == 'nt':
        return os.path.join(os.environ['APPDATA'], 'gcloud')
    else:
        return os.path.join(Path.home(), '.config', 'gcloud')


def get_default_credentials():
    if 'GOOGLE_APPLICATION_CREDENTIALS' in os.environ:
        return os.environ.get('GOOGLE_APPLICATION_CREDENTIALS')
    else:
        return os.path.join(get_auth_directory(), 'application_default_credentials.json')


@attr.s
class GCPDiskTypes(object):
    disk_type_list = [
        {
            "type": 'pd-standard'
        },
        {
            "type": 'pd-balanced'
        },
        {
            "type": 'pd-ssd'
        }
    ]


@attr.s
class GCPImageProjects(object):
    projects = [
        {
            "project": 'centos-cloud',
            "description": "CentOS Linux",
            "os_id": "centos",
            "user": "centos",
            "pattern": "centos-stream-(.+?)-v*"
        },
        {
            "project": 'debian-cloud',
            "description": "Debian Linux",
            "os_id": "debian",
            "user": "admin",
            "pattern": "debian-(.+?)-*-v*"
        },
        {
            "project": 'fedora-cloud',
            "description": "Fedora Linux",
            "os_id": "fedora",
            "user": "fedora",
            "pattern": "fedora-cloud-base-gcp-(.+?)-1-2-x86-64"
        },
        {
            "project": 'opensuse-cloud',
            "description": "OpenSUSE Linux",
            "os_id": "opensuse-leap",
            "user": "admin",
            "pattern": "opensuse-leap-(.+?)-*-v*"
        },
        {
            "project": 'rhel-cloud',
            "description": "Red Hat Enterprise Linux",
            "os_id": "rhel",
            "user": "admin",
            "pattern": "rhel-(.+?)-v*"
        },
        {
            "project": 'rocky-linux-cloud',
            "description": "Rocky Linux",
            "os_id": "rocky",
            "user": "rocky",
            "pattern": "rocky-linux-(.+?)-v*"
        },
        {
            "project": 'suse-cloud',
            "description": "SUSE Linux Enterprise Server",
            "os_id": "sles",
            "user": "admin",
            "pattern": "sles-(.+?)-*-v*"
        },
        {
            "project": 'ubuntu-os-cloud',
            "description": "Ubuntu Linux",
            "os_id": "ubuntu",
            "user": "ubuntu",
            "pattern": r"ubuntu-(.+?)-*-v*"
        },
        {
            "project": 'ubuntu-os-pro-cloud',
            "description": "Ubuntu Pro Linux",
            "os_id": "ubuntu",
            "user": "ubuntu",
            "pattern": r"ubuntu-pro-(.+?)-*-v*"
        },
        {
            "project": 'windows-cloud',
            "description": "Windows Server",
            "os_id": "windows",
            "user": "adminuser",
            "pattern": r"windows-server-(.+?)-dc-v[0-9]+"
        }
    ]


@attr.s
class ComputeTypes(object):
    general_purpose = ['n2', 't2a', 'n2d']
    compute_optimized = ['c2', 'c2d']
    memory_optimized = ['m2', 'm3']

    def as_list(self) -> list:
        flat_list = []
        for element in [self.general_purpose, self.compute_optimized, self.memory_optimized]:
            flat_list.extend(element)
        return flat_list



##
##
import base64
import io
import logging
from couchformation.exception import FatalError
from couchformation.config import get_project_dir, get_base_dir
from couchformation.deployment import NodeGroup, MetadataManager, BuildManager
from couchformation.executor.targets import TargetProfile, ProvisionerProfile, BuildProfile, DeployStrategy, DeployMode, CloudProfileBase
from couchformation.executor.dispatch import JobDispatch
from couchformation.util import FileManager, GenericAttrClass, CloudUtility, dict_merge_not_none

logger = logging.getLogger('couchformation.exec.process')
logger.addHandler(logging.NullHandler())


class ProjectError(FatalError):
    pass


class Project(object):

    def __init__(self, args, remainder):
        self.options = args
        self.remainder = remainder
        self.cloud = self.options.cloud
        self.provisioner = self.options.provisioner
        self.strategy = DeployStrategy()

    def create(self):
        logger.info(f"Creating new service {self.options.name}")
        profile = TargetProfile(self.remainder).get(self.cloud)
        missing = profile.check_required_options()
        if missing:
            raise ProjectError(f"Missing required parameters: {','.join(missing)}")
        BuildManager(self.options, self.remainder).validate()
        NodeGroup(self.options).create_node_group(profile.options)
        MetadataManager(self.options.project).print_services()

    def add(self):
        logger.info(f"Adding node group to service {self.options.name}")
        profile = TargetProfile(self.remainder).get(self.cloud)
        missing = profile.check_required_options()
        if missing:
            raise ProjectError(f"Missing required parameters: {','.join(missing)}")
        BuildManager(self.options, self.remainder).validate()
        NodeGroup(self.options).add_to_node_group(profile.options)
        MetadataManager(self.options.project).print_services()

    def deploy(self, service=None, skip_provision=False):
        password = NodeGroup(self.options).create_credentials()
        private_key, ca_cert = NodeGroup(self.options).create_ca()
        for group in NodeGroup(self.options).get_node_groups():
            self._test_cloud(group)
        for group in NodeGroup(self.options).get_node_groups():
            if service and group[0].get('name') != service:
                continue
            strategy = self.strategy.get(group[0].get('build'))
            cloud = group[0].get('cloud')
            region = group[0].get('region') if group[0].get('region') else "local"
            if strategy.deployer == DeployMode.node.value:
                self._deploy_network(cloud, region)
                self._deploy_node(group, password, private_key, ca_cert, skip_provision)
            elif strategy.deployer == DeployMode.saas.value:
                self._deploy_saas(group, password)

    def destroy(self, service=None):
        for group in NodeGroup(self.options).get_node_groups():
            self._test_cloud(group)
        for group in reversed(list(NodeGroup(self.options).get_node_groups())):
            if service and group[0].get('name') != service:
                continue
            strategy = self.strategy.get(group[0].get('build'))
            cloud = group[0].get('cloud')
            region = group[0].get('region') if group[0].get('region') else "local"
            if strategy.deployer == DeployMode.node.value:
                self._destroy_node(group)
                self._destroy_network(cloud, region)
            elif strategy.deployer == DeployMode.saas.value:
                self._destroy_saas(group)

    def list(self, api=False, service=None):
        return_list = []
        for group in NodeGroup(self.options).get_node_groups():
            self._test_cloud(group)
        for group in NodeGroup(self.options).get_node_groups():
            if service and group[0].get('name') != service:
                continue
            strategy = self.strategy.get(group[0].get('build'))
            if strategy.deployer == DeployMode.node.value:
                results = self._list_node(group, api)
                return_list.extend(results)
            elif strategy.deployer == DeployMode.saas.value:
                results = self._list_saas(group, api)
                return_list.extend(results)
        password = self.credential()
        return_list = [dict(item, project_password=password) for item in return_list]
        if not api:
            logger.info(f"Project Credentials: {password} ")
        if not api and self.options.extended:
            logger.info(f"Project CA:\n{base64.b64decode(MetadataManager(self.options.project).get_project_ca()).decode()} ")
        return return_list

    def accept_peering(self, service=None):
        for group in NodeGroup(self.options).get_node_groups():
            self._test_cloud(group)
        for group in NodeGroup(self.options).get_node_groups():
            if service and group[0].get('name') != service:
                continue
            cloud = group[0].get('cloud')
            region = group[0].get('region') if group[0].get('region') else "local"
            self._accept_peering(cloud, region)

    def credential(self):
        return NodeGroup(self.options).get_credentials()

    def _test_cloud(self, group):
        runner = JobDispatch()
        cloud = group[0].get('cloud')
        profile = TargetProfile(self.remainder).get(cloud)
        base_parameters = dict_merge_not_none(vars(profile.options), group[0].as_dict)
        if not base_parameters.get('region'):
            base_parameters['region'] = CloudUtility.get_default_region(cloud)
        runner.foreground(profile.base.driver, profile.base.module, profile.base.test, base_parameters)

    def _deploy_network(self, cloud, region):
        runner = JobDispatch()
        net = NodeGroup(self.options).get_network(cloud, region)
        profile = TargetProfile(self.remainder).get(cloud)
        module = profile.network.driver
        instance = profile.network.module
        method = profile.network.deploy
        runner.foreground(module, instance, method, net.as_dict)

    @staticmethod
    def peer_network(cloud, parameters):
        runner = JobDispatch()
        profile = CloudProfileBase(cloud).profile
        module = profile.network.driver
        instance = profile.network.module
        method = profile.network.peer
        runner.foreground(module, instance, method, parameters)

    @staticmethod
    def unpeer_network(cloud, parameters):
        runner = JobDispatch()
        profile = CloudProfileBase(cloud).profile
        module = profile.network.driver
        instance = profile.network.module
        method = profile.network.unpeer
        runner.foreground(module, instance, method, parameters)

    def _accept_peering(self, cloud, region):
        runner = JobDispatch()
        net = NodeGroup(self.options).get_network(cloud, region)
        profile = TargetProfile(self.remainder).get(cloud)
        module = profile.network.driver
        instance = profile.network.module
        method = profile.network.peer
        runner.foreground(module, instance, method, profile.merge_options(net.as_dict))

    def _deploy_saas(self, group, password):
        runner = JobDispatch()
        cloud = group[0].get('cloud')
        profile = TargetProfile(self.remainder).get(cloud)
        module = profile.node.driver
        instance = profile.node.module
        compose = profile.node.compose
        deploy = profile.node.deploy

        for n, db in enumerate(group):
            parameters = db.as_dict
            parameters['number'] = n + 1
            runner.foreground(module, instance, compose, parameters)

        main_params = group[0].as_dict
        main_params.update({
            'password': password
        })

        if group[0].get('connect'):
            connect_list = self.list(api=True, service=group[0].get('connect'))
            if len(connect_list) == 0:
                raise ProjectError(f"Connect: No services in {group[0].get('connect')}")
            logger.info(f"Connecting service {group[0].get('name')} to {group[0].get('connect')}")
            main_params.update({
                'instance_id': connect_list[0].get('instance_id'),
                'instance_name': connect_list[0].get('name'),
            })

        runner.foreground(module, instance, deploy, main_params)

    def _deploy_node(self, group, password, private_key, ca_cert, skip_provision=False):
        number = 0
        runner = JobDispatch()

        for db in group:
            cloud = db.get('cloud')
            profile = TargetProfile(self.remainder).get(cloud)
            module = profile.node.driver
            instance = profile.node.module
            method = profile.node.deploy
            quantity = db['quantity'] if db['quantity'] else 1
            for n in range(int(quantity)):
                number += 1
                logger.info(f"Deploying service {db.get('name')} node group {db.get('group')} node {number}")
                parameters = db.as_dict
                parameters['number'] = number
                runner.dispatch(module, instance, method, parameters)
        result_list = list(runner.join())
        if len(result_list) != number:
            raise ProjectError(f"Partial deployment: deployed {len(result_list)} expected {number}")
        result_list = sorted(result_list, key=lambda d: d['name'])
        private_ip_list = [d['private_ip'] for d in result_list]
        public_ip_list = [d['public_ip'] for d in result_list]
        private_host_list = [d['private_hostname'] for d in result_list if d.get('private_hostname')]
        public_host_list = [d['public_hostname'] for d in result_list if d.get('public_hostname')]
        service_list = [d['services'] for d in result_list if d.get('services', 'default')]
        result_list = [dict(item,
                            private_ip_list=private_ip_list,
                            public_ip_list=public_ip_list,
                            private_host_list=private_host_list,
                            public_host_list=public_host_list,
                            service_list=service_list) for item in result_list]
        result_list = [dict(item, password=password) if 'password' not in item else item for item in result_list]

        if skip_provision:
            return

        if group[0].get('connect'):
            connect_list = self.list(api=True, service=group[0].get('connect'))
            connect_list = sorted(connect_list, key=lambda d: d['name'])
            if len(connect_list) == 0:
                raise ProjectError(f"Connect: No nodes in service {group[0].get('connect')}")
            logger.info(f"Connecting service {group[0].get('name')} to {group[0].get('connect')}")
            connect_list = [d['private_ip'] for d in connect_list]
            result_list = [dict(item, connect=connect_list) for item in result_list]

        provisioner_name = ProvisionerProfile().search(group[0])
        if not provisioner_name:
            raise ProjectError("No provisioner matches configuration")

        logger.info(f"Selected provisioner {provisioner_name}")

        if group[0].get('upload'):
            provisioner = ProvisionerProfile().get(provisioner_name)
            p_module = provisioner.driver
            p_instance = provisioner.module
            p_upload = provisioner.upload
            p_list = [provisioner.parameter_gen(result, group[0].as_dict) for result in result_list]
            for p_set in p_list:
                logger.info(f"Uploading file {p_set.get('upload')}")
                runner.dispatch(p_module, p_instance, p_upload, p_set)
            exit_codes = list(runner.join())
            if any(n != 0 for n in exit_codes):
                raise ProjectError(f"Provisioning step failed")

        default_seq = BuildProfile().get('default')

        for build_config in default_seq.get(provisioner_name):
            if group[0].get('os_id') in build_config.exclude:
                continue
            provisioner = ProvisionerProfile().get(build_config.provisioner)
            p_module = provisioner.driver
            p_instance = provisioner.module
            p_method = provisioner.method
            p_list = [provisioner.parameter_gen(result, group[0].as_dict) for result in result_list]
            for step, command in enumerate(build_config.commands):
                for p_set in p_list:
                    logger.info(f"Provisioning node {p_set.get('name')} - default step #{step + 1}")
                    runner.dispatch(p_module, p_instance, p_method, p_set, command, build_config.root)
                exit_codes = list(runner.join())
                if any(n != 0 for n in exit_codes):
                    raise ProjectError(f"Provisioning step failed")

        build_seq = BuildProfile().get(group[0].get('build'))

        for build_config in build_seq.get(provisioner_name):
            if group[0].get('os_id') in build_config.exclude:
                continue
            provisioner = ProvisionerProfile().get(build_config.provisioner)
            p_module = provisioner.driver
            p_instance = provisioner.module
            p_method = provisioner.method
            p_files = provisioner.files
            p_list = [provisioner.parameter_gen(result, group[0].as_dict) for result in result_list]
            for file in build_config.files:
                destination = build_config.files[file]
                if file == 'ca_cert':
                    fl = io.BytesIO(base64.b64decode(ca_cert))
                elif file == 'ca_key':
                    fl = io.BytesIO(base64.b64decode(private_key))
                else:
                    fl = open(file, 'rb')
                logger.info(f"Copying file {file} to {destination}")
                for p_set in p_list:
                    copy_class = runner.get_class(p_module, p_instance, p_set, '', build_config.root)
                    runner.run_method(copy_class, p_files, fl, destination)
            for step, command in enumerate(build_config.commands):
                for p_set in p_list:
                    logger.info(f"Provisioning node {p_set.get('name')} - build step #{step + 1}")
                    runner.dispatch(p_module, p_instance, p_method, p_set, command, build_config.root)
                exit_codes = list(runner.join())
                if any(n != 0 for n in exit_codes):
                    raise ProjectError(f"Provisioning step failed")

    def _destroy_node(self, group):
        number = 0
        runner = JobDispatch()

        for db in group:
            cloud = db.get('cloud')
            profile = TargetProfile(self.remainder).get(cloud)
            module = profile.node.driver
            instance = profile.node.module
            method = profile.node.destroy
            runner.foreground(profile.base.driver, profile.base.module, profile.base.test, db.as_dict)
            quantity = db['quantity'] if db['quantity'] else 1
            for n in range(int(quantity)):
                number += 1
                logger.info(f"Removing service {db.get('name')} node group {db.get('group')} node {number}")
                parameters = db.as_dict
                parameters['number'] = number
                runner.dispatch(module, instance, method, parameters)
        list(runner.join())

    def _destroy_saas(self, group):
        runner = JobDispatch()
        cloud = group[0].get('cloud')
        profile = TargetProfile(self.remainder).get(cloud)
        module = profile.node.driver
        instance = profile.node.module
        method = profile.node.destroy
        runner.foreground(module, instance, method, group[0].as_dict)

    def _destroy_network(self, cloud, region):
        runner = JobDispatch()
        net = NodeGroup(self.options).get_network(cloud, region)
        profile = TargetProfile(self.remainder).get(cloud)
        module = profile.network.driver
        instance = profile.network.module
        method = profile.network.destroy
        runner.dispatch(module, instance, method, net.as_dict)

    def copy(self):
        logger.info(f"Copying project {self.options.project} to {self.options.to}")
        MetadataManager(self.options.project).copy_project(self.options.to)
        MetadataManager(self.options.to).print_services()

    def remove(self):
        if self.options.name:
            service = self.options.name
        else:
            service = None
        logger.info("Removing All Services" if not service else f"Removing {service}")
        self.destroy(service=service)
        NodeGroup(self.options).remove_node_groups(service)
        MetadataManager(self.options.project).print_services()

    def clean(self):
        logger.info("Cleaning project")
        self.destroy()
        NodeGroup(self.options).clean_node_groups()
        NodeGroup(self.options).clean_base()

    def _list_node(self, group, api=False):
        return_list = []
        number = 0
        runner = JobDispatch()

        for db in group:
            cloud = db.get('cloud')
            profile = TargetProfile(self.remainder).get(cloud)
            module = profile.node.driver
            instance = profile.node.module
            method = profile.node.info
            quantity = db['quantity'] if db['quantity'] else 1
            for n in range(int(quantity)):
                number += 1
                parameters = db.as_dict
                parameters['number'] = number
                runner.dispatch(module, instance, method, parameters)
        result_list = list(runner.join())

        if any(d.get('name') is None for d in result_list):
            return return_list

        result_list = sorted(result_list, key=lambda d: d['name'])

        if not api:
            logger.info(f"Service: {group[0].get('name')}")
        for result in result_list:
            if not api:
                output = [f"Node: {result.get('name')}",
                          f"Private IP: {result.get('private_ip'):<15}",
                          f"Public IP: {result.get('public_ip'):<15}"]

                if result.get('host_password'):
                    output.append(f"Password: {result.get('host_password'):<18}")

                if result.get('public_hostname') is not None:
                    output.append(f"DNS Name: {result.get('public_hostname', 'N/A')}")

                output.append(f"Services: {result.get('services')}")

                logger.info(' '.join(output))
            return_list.append(result)

        return return_list

    def _list_saas(self, group, api=False):
        return_list = []
        runner = JobDispatch()

        cloud = group[0].get('cloud')
        profile = TargetProfile(self.remainder).get(cloud)
        module = profile.node.driver
        instance = profile.node.module
        method = profile.node.info

        result = runner.foreground(module, instance, method, group[0].as_dict)
        return_list.append(result)

        if not api:
            logger.info(f"ID: {result.get('instance_id')}")
            logger.info(f"Name: {result.get('name')}")
            logger.info(f"Connect String: {result.get('connect_string')}")
            logger.info(f"Cloud: {result.get('provider')}")
            logger.info(f"Network CIDR: {result.get('cidr')}")
            logger.info(f"Allow CIDR: {result.get('allow')}")
            if result.get('password'):
                logger.info(f"Password: {result.get('password')}")

        return return_list

    @staticmethod
    def list_projects():
        base_path = get_base_dir()
        for project in sorted(list(FileManager().list_dir(base_path))):
            if MetadataManager(project).exists:
                MetadataManager(project).print_services()

    def project_show(self):
        logger.info(f"Displaying project {self.options.project}")
        profile = TargetProfile(self.remainder).get(self.cloud)
        base_path = get_base_dir()
        for project in sorted(list(FileManager().list_dir(base_path))):
            if project != self.options.project:
                continue
            if MetadataManager(project).exists:
                MetadataManager(project).print_project(profile.options)

    def project_cli(self):
        base_path = get_base_dir()
        for project in sorted(list(FileManager().list_dir(base_path))):
            if project != self.options.project:
                continue
            if MetadataManager(project).exists:
                MetadataManager(project).print_cli(self.options)

    def service_edit(self):
        logger.info(f"Updating service {self.options.name}")
        profile = TargetProfile(self.remainder).get(self.cloud)
        unsupported = profile.check_undefined_options()
        if unsupported:
            raise ProjectError(f"Unsupported parameters: {' '.join(unsupported)}")
        base_path = get_base_dir()
        for project in sorted(list(FileManager().list_dir(base_path))):
            if project != self.options.project:
                continue
            if MetadataManager(project).exists:
                MetadataManager(project).edit_service(self.options.name, self.options.group, profile.options)

    def show_help(self):
        profile = TargetProfile(self.remainder).get(self.cloud)
        logger.info(f"Parameters for cloud \"{self.cloud}\":\n")
        profile.parser.print_help()

    @property
    def location(self):
        return get_project_dir(self.options.project)

    def login(self, cloud: str = None):
        if cloud:
            self._test_cloud([GenericAttrClass(vars(self.options))])
            logger.info(f"Cloud {cloud.upper()} authenticated")
            return
        for n, group in enumerate(NodeGroup(self.options).get_node_groups()):
            self._test_cloud(group)
            cloud = group[0].get('cloud')
            logger.info(f"Node group {n + 1}: cloud {cloud.upper()}")
        logger.info(f"Cloud providers authenticated")



##
##

import logging
import botocore.exceptions
from typing import Union, List
from couchformation.aws.driver.base import CloudBase, AWSDriverError, EmptyResultSet
from couchformation.aws.driver.constants import AWSTagStruct, AWSTag

logger = logging.getLogger('couchformation.aws.driver.network')
logger.addHandler(logging.NullHandler())
logging.getLogger("botocore").setLevel(logging.ERROR)
logging.getLogger("urllib3").setLevel(logging.ERROR)


class Network(CloudBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def list(self, name: str = None) -> Union[List[dict], None]:
        vpc_list = []
        vpcs = []
        extra_args = {}

        try:
            while True:
                result = self.ec2_client.describe_vpcs(**extra_args)
                vpcs.extend(result['Vpcs'])
                if 'NextToken' not in result:
                    break
                extra_args['NextToken'] = result['NextToken']
        except Exception as err:
            raise AWSDriverError(f"error getting VPC list: {err}")

        for vpc_entry in vpcs:
            if name:
                vpc_name = self.get_tag("Name", vpc_entry['Tags'])
                if vpc_name != name:
                    continue
            vpc_block = {'cidr': vpc_entry['CidrBlock'],
                         'default': vpc_entry['IsDefault'],
                         'id': vpc_entry['VpcId']}
            vpc_list.append(vpc_block)

        if len(vpc_list) == 0:
            return None
        else:
            return vpc_list

    @property
    def cidr_list(self):
        try:
            for item in self.list():
                yield item['cidr']
        except EmptyResultSet:
            return iter(())

    def create(self, name: str, cidr: str, tags: dict = None) -> str:
        vpc_tag = AWSTagStruct.build("vpc")
        vpc_tag.add(AWSTag("Name", name))
        if tags:
            for k, v in tags.items():
                vpc_tag.add(AWSTag(k, str(v)))
        try:
            result = self.ec2_client.create_vpc(CidrBlock=cidr, TagSpecifications=[vpc_tag.as_dict])
        except Exception as err:
            raise AWSDriverError(f"error creating VPC: {err}")

        return result['Vpc']['VpcId']

    def enable_dns_hostnames(self, vpc_id: str):
        try:
            self.ec2_client.modify_vpc_attribute(VpcId=vpc_id, EnableDnsHostnames={'Value': True})
        except Exception as err:
            raise AWSDriverError(f"error setting VPC DNS hostname option: {err}")

    def delete(self, vpc_id: str) -> None:
        try:
            self.ec2_client.delete_vpc(VpcId=vpc_id)
        except botocore.exceptions.ClientError as err:
            if err.response['Error']['Code'].endswith('NotFound'):
                return
            raise AWSDriverError(f"ClientError: {err}")
        except Exception as err:
            raise AWSDriverError(f"error deleting VPC: {err}")

    def details(self, vpc_id: str) -> Union[dict, None]:
        try:
            result = self.ec2_client.describe_vpcs(VpcIds=[vpc_id])
            vpc_entry = result['Vpcs'][0]
            vpc_block = {'cidr': vpc_entry['CidrBlock'],
                         'default': vpc_entry['IsDefault'],
                         'id': vpc_entry['VpcId']}
            return vpc_block
        except botocore.exceptions.ClientError as err:
            if err.response['Error']['Code'].endswith('NotFound'):
                return None
            raise AWSDriverError(f"ClientError: {err}")
        except Exception as err:
            raise AWSDriverError(f"error getting VPC details: {err}")

    def peering_details(self, vpc_id: str) -> Union[List[dict], None]:
        extra_args = {}
        peers = []
        peer_list = []
        peer_filter = {
            'Name': "accepter-vpc-info.vpc-id",
            'Values': [
                vpc_id,
            ]
        }
        try:
            while True:
                result = self.ec2_client.describe_vpc_peering_connections(Filters=[peer_filter], **extra_args)
                peers.extend(result['VpcPeeringConnections'])
                if 'NextToken' not in result:
                    break
                extra_args['NextToken'] = result['NextToken']
        except Exception as err:
            raise AWSDriverError(f"error getting VPC list: {err}")

        for peer_entry in peers:
            peer_block = {'cidr': peer_entry.get('RequesterVpcInfo', {}).get('CidrBlock'),
                          'status': peer_entry.get('Status', {}).get('Code'),
                          'id': peer_entry['VpcPeeringConnectionId']}
            peer_list.append(peer_block)

        if len(peer_list) == 0:
            return []
        else:
            return peer_list

    def peering_get(self, pcx_id: str):
        extra_args = {}
        peers = []
        try:
            while True:
                result = self.ec2_client.describe_vpc_peering_connections(VpcPeeringConnectionIds=[pcx_id], **extra_args)
                peers.extend(result['VpcPeeringConnections'])
                if 'NextToken' not in result:
                    break
                extra_args['NextToken'] = result['NextToken']
        except botocore.exceptions.ClientError as err:
            if err.response['Error']['Code'].endswith('NotFound'):
                return None
            raise AWSDriverError(f"ClientError: {err}")
        except Exception as err:
            raise AWSDriverError(f"error getting VPC list: {err}")

        if len(peers) == 0:
            return None
        else:
            peer_block = {'cidr': peers[0].get('RequesterVpcInfo', {}).get('CidrBlock'),
                          'status': peers[0].get('Status', {}).get('Code'),
                          'id': peers[0]['VpcPeeringConnectionId']}
            return peer_block

    def peering_accept(self, peer_id: str):
        try:
            self.ec2_client.accept_vpc_peering_connection(VpcPeeringConnectionId=peer_id)
        except botocore.exceptions.ClientError as err:
            raise AWSDriverError(f"ClientError: {err}")
        except Exception as err:
            raise AWSDriverError(f"error accepting peering request: {err}")

    def peering_delete(self, peer_id: str):
        try:
            self.ec2_client.delete_vpc_peering_connection(VpcPeeringConnectionId=peer_id)
        except botocore.exceptions.ClientError as err:
            raise AWSDriverError(f"ClientError: {err}")
        except Exception as err:
            raise AWSDriverError(f"error accepting peering request: {err}")


class Subnet(CloudBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger = logging.getLogger(self.__class__.__name__)

    def list(self, vpc_id: str, zone: Union[str, None] = None, filter_keys_exist: Union[List[str], None] = None) -> List[dict]:
        subnet_list = []
        subnets = []
        extra_args = {}
        subnet_filter = [
            {
                'Name': 'vpc-id',
                'Values': [
                    vpc_id,
                ]
            }
        ]

        if zone:
            subnet_filter.append(
                {
                    'Name': 'availability-zone',
                    'Values': [
                        zone,
                    ]
                }
            )

        try:
            while True:
                result = self.ec2_client.describe_subnets(**extra_args, Filters=subnet_filter)
                subnets.extend(result['Subnets'])
                if 'NextToken' not in result:
                    break
                extra_args['NextToken'] = result['NextToken']
        except Exception as err:
            raise AWSDriverError(f"error getting subnets: {err}")

        for subnet in subnets:
            net_block = {'cidr': subnet['CidrBlock'],
                         'name': subnet['SubnetId'],
                         'vpc': subnet['VpcId'],
                         'zone': subnet['AvailabilityZone'],
                         'default': subnet['DefaultForAz'],
                         'public': subnet['MapPublicIpOnLaunch']}
            if filter_keys_exist:
                if not all(key in net_block for key in filter_keys_exist):
                    continue
            subnet_list.append(net_block)

        if len(subnet_list) == 0:
            raise EmptyResultSet(f"no subnets found")

        return subnet_list

    def create(self, name: str, vpc_id: str, zone: str, cidr: str, tags: dict = None) -> str:
        subnet_tag = AWSTagStruct.build("subnet")
        subnet_tag.add(AWSTag("Name", name))
        if tags:
            for k, v in tags.items():
                subnet_tag.add(AWSTag(k, str(v)))
        try:
            result = self.ec2_client.create_subnet(VpcId=vpc_id, AvailabilityZone=zone, CidrBlock=cidr, TagSpecifications=[subnet_tag.as_dict])
            subnet_id = result['Subnet']['SubnetId']
            self.ec2_client.modify_subnet_attribute(SubnetId=subnet_id, MapPublicIpOnLaunch={'Value': True})
            return subnet_id
        except Exception as err:
            AWSDriverError(f"error creating subnet: {err}")

    def details(self, subnet_id: str) -> Union[dict, None]:
        try:
            result = self.ec2_client.describe_subnets(SubnetIds=[subnet_id])
            subnet = result['Subnets'][0]
            net_block = {'cidr': subnet['CidrBlock'],
                         'name': subnet['SubnetId'],
                         'vpc': subnet['VpcId'],
                         'zone': subnet['AvailabilityZone'],
                         'default': subnet['DefaultForAz'],
                         'public': subnet['MapPublicIpOnLaunch']}
            return net_block
        except IndexError:
            return None
        except botocore.exceptions.ClientError as err:
            if err.response['Error']['Code'].endswith('NotFound'):
                return None
            raise AWSDriverError(f"ClientError: {err}")
        except Exception as err:
            raise AWSDriverError(f"error getting VPC details: {err}")

    def delete(self, subnet_id: str) -> None:
        try:
            self.ec2_client.delete_subnet(SubnetId=subnet_id)
        except botocore.exceptions.ClientError as err:
            if err.response['Error']['Code'].endswith('NotFound'):
                return
            raise AWSDriverError(f"ClientError: {err}")
        except Exception as err:
            raise AWSDriverError(f"error deleting subnet: {err}")



##
##

import logging
import botocore.exceptions
from couchformation.aws.driver.base import CloudBase, AWSDriverError

logger = logging.getLogger('couchformation.aws.driver.cost')
logger.addHandler(logging.NullHandler())
logging.getLogger("botocore").setLevel(logging.ERROR)
logging.getLogger("urllib3").setLevel(logging.ERROR)


class Cost(CloudBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_cost(self, machine_type: str, region: str):
        filters = [
            {
                'Type': 'TERM_MATCH',
                'Field': 'instanceType',
                'Value': machine_type
            }
        ]
        try:
            result = self.cost_client.get_attribute_values(ServiceCode='AmazonEC2', AttributeName='instanceType')
            print(result)
            result = self.cost_client.get_products(ServiceCode='AmazonEC2', Filters=filters)
            print(result)
        except botocore.exceptions.ClientError as err:
            if err.response['Error']['Code'].endswith('AccessDeniedException'):
                return None
            raise AWSDriverError(f"ClientError: {err}")
        except Exception as err:
            raise AWSDriverError(f"error getting cost for machine type: {err}")



##
##

import logging
from couchformation.aws.driver.base import CloudBase, AWSDriverError, EmptyResultSet
from couchformation.aws.driver.constants import ComputeTypes
import couchformation.constants as C

logger = logging.getLogger('couchformation.aws.driver.machine')
logger.addHandler(logging.NullHandler())
logging.getLogger("botocore").setLevel(logging.ERROR)
logging.getLogger("urllib3").setLevel(logging.ERROR)


class MachineType(CloudBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def list(self, architecture: str = 'x86_64') -> list:
        type_list = []
        types = []
        filters = []
        prefix_list = []
        extra_args = {}

        for prefix in ComputeTypes().as_list():
            prefix_list.append(f"{prefix}.*")

        filters.append(
            {
                'Name': 'instance-type',
                'Values': prefix_list
            }
        )
        filters.append(
            {
                'Name': 'processor-info.supported-architecture',
                'Values': [architecture]
            }
        )

        try:
            while True:
                result = self.ec2_client.describe_instance_types(Filters=filters, **extra_args)
                types.extend(result['InstanceTypes'])
                if 'NextToken' not in result:
                    break
                extra_args['NextToken'] = result['NextToken']
        except Exception as err:
            raise AWSDriverError(f"error getting instance types: {err}")

        for machine in types:
            key_block = {'name': machine['InstanceType'],
                         'cpu': int(machine['VCpuInfo']['DefaultVCpus']),
                         'memory': int(machine['MemoryInfo']['SizeInMiB']),
                         'arch': machine.get('ProcessorInfo', {}).get('SupportedArchitectures'),
                         'clock': machine.get('ProcessorInfo', {}).get('SustainedClockSpeedInGhz'),
                         'network': machine.get('NetworkInfo', {}).get('NetworkPerformance'),
                         'nvme': machine.get('EbsInfo', {}).get('NvmeSupport'),
                         'hypervisor': machine.get('Hypervisor')}
            type_list.append(key_block)

        if len(type_list) == 0:
            raise EmptyResultSet(f"no instance types found")

        return type_list

    def get_machine_zones(self, instance_type: str):
        machine_filter = [
            {
                'Name': 'instance-type',
                'Values': [
                    instance_type,
                ]
            }
        ]
        try:
            result = self.ec2_client.describe_instance_type_offerings(LocationType='availability-zone', Filters=machine_filter)
            return list(loc['Location'] for loc in result.get('InstanceTypeOfferings', []))
        except Exception as err:
            raise AWSDriverError(f"error getting instance type details: {err}")

    def get_machine_types(self, architecture: str = 'x86_64'):
        result_list = []
        machine_list = self.list(architecture)
        machine_list = sorted(machine_list, key=lambda m: m['name'])

        for machine_type in C.MACHINE_TYPES:
            machine = next((m for m in machine_list if m['cpu'] == machine_type['cpu'] and m['memory'] == machine_type['memory']), None)
            if not machine:
                continue
            machine.update(dict(machine_type=machine_type['name']))
            result_list.append(machine)

        return result_list

    def get_machine(self, name: str, architecture: str = 'x86_64'):
        machine_list = self.get_machine_types(architecture)
        return next((m for m in machine_list if m['machine_type'] == name), None)

    def get_next_machine(self, name: str, architecture: str = 'x86_64'):
        machine_list = self.get_machine_types(architecture)
        index = next((i for (i, d) in enumerate(machine_list) if d['name'] == name), 0)
        return machine_list[index + 1]

    def details(self, instance_type: str) -> dict:
        try:
            result = self.ec2_client.describe_instance_types(InstanceTypes=[instance_type])
        except Exception as err:
            raise AWSDriverError(f"error getting instance type details: {err}")

        if len(result['InstanceTypes']) == 0:
            raise EmptyResultSet(f"can not find instance type {instance_type}")

        machine = result['InstanceTypes'][0]

        return {'name': machine['InstanceType'],
                'cpu': int(machine['VCpuInfo']['DefaultVCpus']),
                'memory': int(machine['MemoryInfo']['SizeInMiB']),
                'arch': machine.get('ProcessorInfo', {}).get('SupportedArchitectures'),
                'clock': machine.get('ProcessorInfo', {}).get('SustainedClockSpeedInGhz'),
                'network': machine.get('NetworkInfo', {}).get('NetworkPerformance'),
                'hypervisor': machine.get('Hypervisor')}



##
##

import sys
import os
import inspect
import logging
import logging.handlers
import traceback
import couchformation.constants as C
from datetime import datetime
from couchformation.config import get_log_dir

logger = logging.getLogger('couchformation.error')
logger.addHandler(logging.NullHandler())


class CustomCrashFormatter(logging.Formatter):
    FORMATS = {
        logging.DEBUG: f"{C.FORMAT_MESSAGE}",
        logging.INFO: f"{C.FORMAT_MESSAGE}",
        logging.WARNING: f"{C.FORMAT_MESSAGE}",
        logging.ERROR: f"{C.FORMAT_MESSAGE}",
        logging.CRITICAL: f"{C.FORMAT_MESSAGE}"
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        if logging.DEBUG >= logging.root.level:
            log_fmt += C.FORMAT_EXTRA
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


class FatalError(Exception):

    def __init__(self, message):
        frame = inspect.currentframe().f_back
        (filename, line, function, lines, index) = inspect.getframeinfo(frame)
        filename = os.path.basename(filename)
        logging.debug(f"Error: {type(self).__name__} in {filename} {function} at line {line}: {message}")
        logging.error(f"{message} [{filename}:{line}]")

        crash_log_file = os.path.join(get_log_dir(), "crash.log")
        logging.debug(f"See {crash_log_file} for stack trace")

        with open(crash_log_file, 'a') as log:
            now = datetime.now()
            time_string = now.strftime("%D %I:%M:%S %p")
            log.write(f"---- BEGIN {time_string} ----\n")
            log.write(f"== <ERROR> ==\n")
            log.write(f"{message} [{filename}:{line}]\n")
            trace_output = traceback.format_exc()
            if trace_output:
                log.write(f"== <TRACE> ==\n")
                log.write(trace_output)
            log.write(f"---- END ----\n")
            log.flush()

        sys.exit(1)


class NonFatalLogError(Exception):

    def __init__(self, message):
        frame = inspect.currentframe().f_back
        (filename, line, function, lines, index) = inspect.getframeinfo(frame)
        filename = os.path.basename(filename)
        logging.debug(f"Error: {type(self).__name__} in {filename} {function} at line {line}: {message}")
        logging.debug(''.join(traceback.format_stack(frame)))
        logging.error(f"{message} [{filename}:{line}]")


class NonFatalError(Exception):

    def __init__(self, message):
        frame = inspect.currentframe().f_back
        (filename, line, function, lines, index) = inspect.getframeinfo(frame)
        filename = os.path.basename(filename)
        self.message = f"Error: {type(self).__name__} in {filename} {function} at line {line}: {message}"
        super().__init__(self.message)



##
##

import re
import logging
from itertools import cycle, islice
from couchformation.azure.driver.base import CloudBase
from couchformation.azure.driver.network import Network
from couchformation.azure.driver.instance import Instance
from couchformation.azure.driver.machine import MachineType
from couchformation.azure.driver.disk import Disk
from couchformation.azure.driver.image import Image
from couchformation.azure.driver.dns import DNS
from couchformation.azure.driver.private_dns import PrivateDNS
from couchformation.azure.network import AzureNetwork
from couchformation.deployment import MetadataManager
from couchformation.config import get_state_file, get_state_dir, PortSettingSet, State
from couchformation.ssh import SSHUtil
from couchformation.exception import FatalError
from couchformation.kvdb import KeyValueStore
from couchformation.util import FileManager, Synchronize, UUIDGen
from couchformation.util import PasswordUtility

logger = logging.getLogger('couchformation.azure.node')
logger.addHandler(logging.NullHandler())


class AzureNodeError(FatalError):
    pass


class AzureDeployment(object):

    def __init__(self, parameters: dict):
        self.parameters = parameters
        self.name = parameters.get('name')
        self.project = parameters.get('project')
        self.build = parameters.get('build')
        self.region = parameters.get('region')
        self.zone = parameters.get('zone')
        self.auth_mode = parameters.get('auth_mode')
        self.profile = parameters.get('profile')
        self.ssh_key = parameters.get('ssh_key')
        self.os_id = parameters.get('os_id')
        self.os_version = parameters.get('os_version')
        self.feature = parameters.get('feature')
        self.cloud = parameters.get('cloud')
        self.group = parameters.get('group')
        self.number = parameters.get('number')
        self.machine_type = parameters.get('machine_type')
        self.ports = parameters.get('ports')
        self.ultra = parameters.get('ultra') if parameters.get('ultra') else False
        self.password = parameters.get('password') if parameters.get('password') else PasswordUtility().generate(16)
        self.volume_size = parameters.get('volume_size') if parameters.get('volume_size') else "256"
        self.services = parameters.get('services') if parameters.get('services') else "default"

        project_uid = MetadataManager(self.project).project_uid
        self.asset_prefix = f"cf-{project_uid}"
        self.rg_name = f"{self.asset_prefix}-rg"
        self.node_name = f"{self.name}-node-{self.number:02d}"
        self.boot_disk = f"{self.name}-boot-{self.number:02d}"
        self.swap_disk = f"{self.name}-swap-{self.number:02d}"
        self.data_disk = f"{self.name}-data-{self.number:02d}"
        self.node_pub_ip = f"{self.name}-node-{self.number:02d}-pub-ip"
        self.node_nic = f"{self.name}-node-{self.number:02d}-nic"
        node_code = UUIDGen().text_hash(self.node_name)
        self.node_encoded = f"{self.asset_prefix}-{node_code}-node"
        self.boot_encoded = f"{self.asset_prefix}-{node_code}-boot"
        self.swap_encoded = f"{self.asset_prefix}-{node_code}-swap"
        self.data_encoded = f"{self.asset_prefix}-{node_code}-data"
        self.pub_ip_encoded = f"{self.asset_prefix}-{node_code}-pub-ip"
        self.nic_encoded = f"{self.asset_prefix}-{node_code}-nic"

        filename = get_state_file(self.project, self.name)

        with Synchronize():
            try:
                state_dir = get_state_dir(self.project, self.name)
                FileManager().make_dir(state_dir)
            except Exception as err:
                raise AzureNodeError(f"can not create state dir: {err}")

        document = self.node_name
        self.state = KeyValueStore(filename, document)

        self.az_network = AzureNetwork(self.parameters)
        self.az_base = CloudBase(self.parameters)

    def check_state(self):
        if self.state.get('resource_group'):
            rg_name = self.state.get('resource_group')
        else:
            rg_name = f"{self.project}-rg"

        if self.state.get('instance_id'):
            result = Instance(self.parameters).details(self.state['instance_id'], rg_name)
            if result is None:
                logger.warning(f"Removing stale state entry for instance {self.state['instance_id']}")
                del self.state['instance_id']
        if self.state.get('node_nic'):
            result = Network(self.parameters).describe_nic(self.state['node_nic'], rg_name)
            if result is None:
                logger.warning(f"Removing stale state entry for NIC {self.state['node_nic']}")
                del self.state['node_nic']
        if self.state.get('node_pub_ip'):
            result = Network(self.parameters).describe_pub_ip(self.state['node_pub_ip'], rg_name)
            if result is None:
                logger.warning(f"Removing stale state entry for public IP {self.state['node_pub_ip']}")
                del self.state['node_pub_ip']
        if self.state.get('swap_disk'):
            result = Disk(self.parameters).details(self.state['swap_disk'], rg_name)
            if result is None:
                logger.warning(f"Removing stale state entry for disk {self.state['swap_disk']}")
                del self.state['swap_disk']
        if self.state.get('data_disk'):
            result = Disk(self.parameters).details(self.state['data_disk'], rg_name)
            if result is None:
                logger.warning(f"Removing stale state entry for disk {self.state['data_disk']}")
                del self.state['data_disk']
        if self.state.get('boot_disk'):
            result = Disk(self.parameters).details(self.state['boot_disk'], rg_name)
            if result is None:
                logger.warning(f"Removing stale state entry for disk {self.state['boot_disk']}")
                del self.state['boot_disk']

    def deploy(self):
        self.check_state()
        subnet_list = []

        self.state['state'] = State.DEPLOYING.value

        if self.state.get('instance_id'):
            logger.info(f"Node {self.node_name} already exists")
            return self.state.as_dict

        ssh_pub_key_text = SSHUtil().get_ssh_public_key(self.ssh_key)
        rg_name = self.az_network.resource_group
        azure_location = self.az_base.region

        for n, zone_state in enumerate(self.az_network.zones):
            subnet_list.append(dict(
                subnet_id=zone_state[2],
                zone=zone_state[0],
                subnet=zone_state[1],
            ))

        if len(subnet_list) == 0:
            raise AzureNodeError(f"can not get subnet list, check project settings")

        if not self.zone:
            subnet_cycle = cycle(subnet_list)
            subnet = next(islice(subnet_cycle, self.number - 1, None))
        else:
            subnet = next((z for z in subnet_list if z['zone'] == self.zone), None)

        if not subnet:
            raise AzureNodeError(f"Can not determine availability zone (check project settings)")

        image = Image(self.parameters).list_standard(os_id=self.os_id, os_version=self.os_version)
        if not image:
            raise AzureNodeError(f"can not find image for type {self.os_id} {self.os_version}")

        logger.info(f"Using image {image['publisher']}/{image['offer']}/{image['sku']} type {image['os_id']} version {image['os_version']}")

        self.state['service'] = self.name
        self.state['username'] = image['os_user']

        machine_type = self.machine_type
        volume_size = self.volume_size
        services = self.services

        if self.feature == "vmp":
            logger.info(f"Enabling nested virtualization")
            virtualization = True
        else:
            virtualization = False

        machine = MachineType(self.parameters).get_machine(self.machine_type, azure_location, virtualization)
        if not machine:
            raise AzureNodeError(f"can not find machine for type {machine_type}")
        machine_name = machine['name']
        machine_ram = int(machine['memory'] / 1024)
        logger.info(f"Selecting machine type {machine_name}")

        if self.ports:
            self.az_network.create_node_group_sg(self.name, self.group, self.ports.split(','))
            logger.info("Requesting service group firewall rule")

        build_ports = PortSettingSet().create().get(self.build)
        if build_ports:
            self.az_network.create_build_sg(self.build)
            logger.info(f"Requesting build {self.build} firewall rule")

        if image['os_id'] == 'windows':
            self.az_network.create_win_sg()
            logger.info("Requesting windows firewall rule")

        logger.info(f"Creating disk {self.swap_encoded} ({self.swap_disk})")
        swap_resource = Disk(self.parameters).create(rg_name, azure_location, subnet['zone'], machine_ram, self.swap_encoded, self.ultra)
        self.state['swap_disk'] = self.swap_encoded

        logger.info(f"Creating disk {self.data_encoded} ({self.data_disk})")
        data_resource = Disk(self.parameters).create(rg_name, azure_location, subnet['zone'], volume_size, self.data_encoded, self.ultra)
        self.state['data_disk'] = self.data_encoded

        logger.info(f"Creating public IP {self.pub_ip_encoded} ({self.node_pub_ip})")
        pub_ip_resource = Network(self.parameters).create_pub_ip(self.pub_ip_encoded, rg_name)
        self.state['node_pub_ip'] = self.pub_ip_encoded

        logger.info(f"Creating NIC {self.nic_encoded} ({self.node_nic})")
        nic_resource = Network(self.parameters).create_nic(self.nic_encoded, subnet['subnet_id'], subnet['zone'], pub_ip_resource.id, rg_name)
        self.state['node_nic'] = self.nic_encoded

        if image['os_id'] == 'windows' and not self.state['host_password']:
            self.state['host_password'] = self.password
        elif self.state['host_password']:
            self.password = self.state['host_password']

        logger.info(f"Creating node {self.node_encoded} ({self.node_name})")
        Instance(self.parameters).run(self.node_encoded,
                                      image['publisher'],
                                      image['offer'],
                                      image['sku'],
                                      subnet['zone'],
                                      nic_resource.id,
                                      image['os_user'],
                                      ssh_pub_key_text,
                                      rg_name,
                                      self.boot_encoded,
                                      machine_type=machine_name,
                                      password=self.password,
                                      ultra=self.ultra)

        logger.info(f"Attaching disk {self.swap_encoded}")
        Instance(self.parameters).attach_disk(self.node_encoded, self.az_base.disk_caching(machine_ram, self.ultra), "1", swap_resource.id, rg_name)
        logger.info(f"Attaching disk {self.data_encoded}")
        Instance(self.parameters).attach_disk(self.node_encoded, self.az_base.disk_caching(volume_size, self.ultra), "2", data_resource.id, rg_name)

        self.state['instance_id'] = self.node_encoded
        self.state['name'] = self.node_encoded
        self.state['services'] = services
        self.state['zone'] = subnet['zone']
        self.state['resource_group'] = rg_name
        self.state['boot_disk'] = self.boot_encoded
        self.az_network.add_service(self.node_name)

        nic_details = Network(self.parameters).describe_nic(self.nic_encoded, rg_name)
        pub_ip_details = Network(self.parameters).describe_pub_ip(self.pub_ip_encoded, rg_name)
        self.state['public_ip'] = pub_ip_details.ip_address
        self.state['private_ip'] = nic_details.ip_configurations[0].private_ip_address

        if self.az_network.public_zone and self.az_network.domain_name and not self.state.get('public_hostname'):
            host_name = f"{self.node_name}.{self.az_network.domain_name}"
            DNS(self.parameters).add_record(self.az_network.public_zone, host_name, [self.state['public_ip']], rg_name)
            self.state['public_zone_id'] = self.az_network.public_zone
            self.state['public_hostname'] = host_name

        if self.az_network.private_zone and self.az_network.domain_name and not self.state.get('private_hostname'):
            host_name = f"{self.node_name}.{self.az_network.domain_name}"
            PrivateDNS(self.parameters).add_record(self.az_network.private_zone, host_name, [self.state['private_ip']], rg_name)
            self.state['private_zone_id'] = self.az_network.private_zone
            self.state['private_hostname'] = host_name

        logger.info(f"Created instance {self.node_name}")
        self.state['state'] = State.DEPLOYED.value
        return self.state.as_dict

    def destroy(self):
        self.state['state'] = State.DESTROYING.value
        rg_name = self.state['resource_group']
        if not rg_name:
            rg_name = self.rg_name
        if self.state.get('public_hostname'):
            domain_id = self.state['public_zone_id']
            name = self.state['public_hostname']
            ip = self.state['public_ip']
            DNS(self.parameters).delete_record(domain_id, name, rg_name)
            logger.info(f"Deleted DNS record for {ip}")
        if self.state.get('private_hostname'):
            domain_id = self.state['private_zone_id']
            name = self.state['private_hostname']
            ip = self.state['private_ip']
            PrivateDNS(self.parameters).delete_record(domain_id, name, rg_name)
            logger.info(f"Deleted DNS record for {ip}")
        if not self.state.get('instance_id'):
            result = Instance(self.parameters).details(self.node_encoded, rg_name)
            if result:
                logger.warning(f"Found orphaned instance {self.node_encoded} - repairing configuration")
                self.state['instance_id'] = self.node_encoded
        if not self.state.get('boot_disk'):
            result = Disk(self.parameters).details(self.boot_encoded, rg_name)
            if result:
                logger.warning(f"Found orphaned boot disk {self.boot_encoded} - repairing configuration")
                self.state['boot_disk'] = self.boot_encoded
        if self.state.get('instance_id'):
            instance_name = self.state['instance_id']
            node_nic = self.state['node_nic']
            node_pub_ip = self.state['node_pub_ip']
            Instance(self.parameters).terminate(instance_name, rg_name)
            logger.info(f"Removed instance {instance_name}")
            Network(self.parameters).delete_nic(node_nic, rg_name)
            logger.info(f"Removed NIC {node_nic}")
            Network(self.parameters).delete_pub_ip(node_pub_ip, rg_name)
            logger.info(f"Removed public IP {node_pub_ip}")
            Disk(self.parameters).delete(self.state['swap_disk'], rg_name)
            logger.info(f"Removed disk {self.state['swap_disk']}")
            Disk(self.parameters).delete(self.state['data_disk'], rg_name)
            logger.info(f"Removed disk {self.state['data_disk']}")
            Disk(self.parameters).delete(self.state['boot_disk'], rg_name)
            logger.info(f"Removed disk {self.state['boot_disk']}")
            self.state.clear()
            self.az_network.remove_service(self.node_name)
            logger.info(f"Removed instance {instance_name}")
            self.state['state'] = State.IDLE.value

    def info(self):
        return self.state.as_dict

    @staticmethod
    def _name_check(value):
        p = re.compile(r"^[a-z]([-_a-z0-9]*[a-z0-9])?$")
        if p.match(value):
            return value
        else:
            raise AzureNodeError("names must only contain letters, numbers, dashes and underscores")



##
##

import logging
from couchformation.exception import FatalError
from couchformation.resources.config_manager import ConfigurationManager
from libcapella.config import CapellaConfig
from libcapella.organization import CapellaOrganization
from libcapella.project import CapellaProject
from libcapella.logic.project import CapellaProjectBuilder

logger = logging.getLogger('couchformation.capella.driver.base')
logger.addHandler(logging.NullHandler())
logging.getLogger("restfull").setLevel(logging.ERROR)


class CapellaDriverError(FatalError):
    pass


class CloudBase(object):

    def __init__(self, parameters: dict):
        self.parameters = parameters
        self._token = None
        self._account_email = None
        self._account_id = None
        self._project_name = None

        cm = ConfigurationManager()
        if cm.get('capella.token'):
            self._token = cm.get('capella.token')
        if cm.get('capella.user'):
            self._account_email = cm.get('capella.user')
        if cm.get('capella.user.id'):
            self._account_id = cm.get('capella.user.id')
        if cm.get('capella.project'):
            self._project_name = cm.get('capella.project')
        else:
            self._project_name = parameters.get('project')

        try:
            if self._token and (self._account_email or self._account_id):
                config_dict = {
                    "token": self._token,
                    "account_email": self._account_email,
                    "project_name": self._project_name,
                    "account_id": self._account_id,
                }
                logger.debug(f"Capella config parameters: {config_dict}")
                config = CapellaConfig(config_dict=config_dict)
            else:
                profile = parameters.get('profile')
                logger.debug(f"Capella credential profile: {profile}")
                config = CapellaConfig(profile=profile)

            self._account_email = config.config.account_email

            if not self._account_email:
                raise CapellaDriverError("Capella account email not set")

            if not config.config.token:
                raise CapellaDriverError("Capella v4 API token not set")

            self.org = CapellaOrganization(config)
            self._project = CapellaProject(self.org, self._project_name, self._account_email)
            if not self._project.id:
                logger.info(f"Creating project {self._project_name}")
                builder = CapellaProjectBuilder()
                builder = builder.name(self._project_name)
                config = builder.build()
                self._project.create(config)
        except Exception as err:
            raise CapellaDriverError(f"can not access Capella project {self._project_name}: {err}")

    def test_session(self):
        try:
            self.org.list()
        except Exception as err:
            raise CapellaDriverError(f"not authorized: {err}")

    @property
    def project_id(self):
        return self._project.id

    @property
    def organization_id(self):
        return self.org.id

    @property
    def project_name(self):
        return self.project_name

    @property
    def project(self):
        return self._project



##
##

import os.path
import socket
import logging
import json
import time
import base64
import sqlite3
import googleapiclient.discovery
import googleapiclient.errors
import google.auth
import google.auth.transport.requests
from pathlib import Path
from google.cloud import storage
from google.oauth2 import service_account
from google.cloud import resourcemanager_v3
from couchformation.config import AuthMode
from couchformation.exception import FatalError, NonFatalError
from couchformation.retry import retry
from couchformation.gcp.driver.constants import get_auth_directory, get_default_credentials

logger = logging.getLogger('couchformation.gcp.driver.base')
logger.addHandler(logging.NullHandler())
logging.getLogger("googleapiclient").setLevel(logging.ERROR)


class GCPDriverError(FatalError):
    pass


class GCPDriverTransientError(NonFatalError):
    pass


class EmptyResultSet(NonFatalError):
    pass


class CloudBase(object):
    cache = {}

    def __init__(self, parameters: dict):
        self.parameters = parameters
        self.auth_directory = get_auth_directory()
        self.gcp_account = None
        self.gcp_project = None
        self.gcp_region = None
        self.gcp_account_file = None
        self._service_account_email = None
        self._user_account_email = None
        self.gcp_zone_list = []
        self.gcp_zone = None

        socket.setdefaulttimeout(120)

        if not parameters.get('auth_mode') or AuthMode[parameters.get('auth_mode')] == AuthMode.default:
            if not CloudBase.cache.get('credentials') or not CloudBase.cache.get('gcp_project'):
                self.credentials, self.gcp_project = self.default_auth()
                CloudBase.cache['credentials'] = self.credentials
                CloudBase.cache['gcp_project'] = self.gcp_project
            else:
                self.credentials = CloudBase.cache.get('credentials')
                self.gcp_project = CloudBase.cache.get('gcp_project')

            if not CloudBase.cache.get('service_account_email') or not CloudBase.cache.get('user_account_email'):
                self._service_account_email, self._user_account_email = self.get_account_email()
                CloudBase.cache['service_account_email'] = self._service_account_email
                CloudBase.cache['user_account_email'] = self._user_account_email
            else:
                self._service_account_email = CloudBase.cache.get('service_account_email')
                self._user_account_email = CloudBase.cache.get('user_account_email')
        elif AuthMode[parameters.get('auth_mode')] == AuthMode.file:
            self.credentials, self.gcp_project, self._service_account_email = self.file_auth()
        else:
            raise GCPDriverError(f"Unsupported auth mode {parameters.get('auth_mode')}")

        self.gcp_client = googleapiclient.discovery.build('compute', 'v1', credentials=self.credentials)
        self.dns_client = googleapiclient.discovery.build('dns', 'v1', credentials=self.credentials)

        if not self.gcp_project:
            raise GCPDriverError(f"can not determine GCP project")

        self.gcp_region = parameters.get('region')

        if not self.gcp_region:
            raise GCPDriverError("region not specified")

        try:
            self.zones()
        except GCPDriverTransientError:
            raise GCPDriverError(f"There is likely an auth config or firewall problem - make sure you can access the GCP API and use \"gcloud auth\" to configure access")
        except Exception as err:
            raise GCPDriverError(f"error getting availability zones: {err}")

    def test_session(self):
        try:
            storage_client = storage.Client(project=self.gcp_project, credentials=self.credentials)
            storage_client.list_buckets()
        except Exception as err:
            raise GCPDriverError(f"not authorized: {err}")

    @staticmethod
    def default_auth():
        try:
            credentials, project_id = google.auth.default()
            return credentials, project_id
        except Exception as err:
            raise GCPDriverError(f"error connecting to GCP: {err}")

    def get_account_email(self):
        try:
            if hasattr(self.credentials, "service_account_email"):
                service_account_email = self.credentials.service_account_email
                account_email = None
            elif hasattr(self.credentials, "signer_email"):
                service_account_email = self.credentials.signer_email
                account_email = None
            else:
                service_account_email = None
                request = google.auth.transport.requests.Request()
                self.credentials.refresh(request=request)
                token_payload = self.credentials.id_token.split('.')[1]
                input_bytes = token_payload.encode('utf-8')
                rem = len(input_bytes) % 4
                if rem > 0:
                    input_bytes += b"=" * (4 - rem)
                json_data = base64.urlsafe_b64decode(input_bytes).decode('utf-8')
                token_data = json.loads(json_data)
                account_email = token_data.get('email')
            return service_account_email, account_email
        except Exception as err:
            raise GCPDriverError(f"error getting GCP account email: {err}")

    @staticmethod
    def get_config_dir():
        if 'CLOUDSDK_CONFIG' in os.environ:
            return os.environ['CLOUDSDK_CONFIG']
        if os.name != 'nt':
            return os.path.join(Path.home(), '.config', 'gcloud')
        if 'APPDATA' in os.environ:
            return os.path.join(os.environ['APPDATA'], 'gcloud')
        drive = os.environ.get('SystemDrive', 'C:')
        return os.path.join(drive, os.path.sep, 'gcloud')

    def get_account(self, account: str):
        account_db = os.path.join(self.get_config_dir(), 'credentials.db')
        connection = sqlite3.connect(
            account_db,
            detect_types=sqlite3.PARSE_DECLTYPES,
            isolation_level=None,
            check_same_thread=True
        )

        cursor = connection.cursor()
        table = cursor.execute('SELECT account_id, value FROM credentials').fetchall()
        for row in table:
            account_id, cred_json = row[0], row[1]
            if account_id == account:
                return json.loads(cred_json)

        return None

    def sa_auth(self, service_account_email):
        auth_data = self.get_account(service_account_email)
        if not auth_data:
            raise GCPDriverError(f"Account {service_account_email} is not configured. Use gcloud auth to add the account.")
        credentials, _ = google.auth.load_credentials_from_dict(auth_data)
        return credentials

    @property
    def project_number(self):
        rm = resourcemanager_v3.ProjectsClient()
        req = resourcemanager_v3.GetProjectRequest(dict(name=f"projects/{self.gcp_project}"))
        res = rm.get_project(request=req)
        project_number = res.name.split('/')[1]
        return project_number

    @property
    def default_sa(self):
        return f"{self.project_number}-compute@developer.gserviceaccount.com"

    def file_auth(self):
        auth_file = get_default_credentials()
        if os.path.exists(auth_file):
            credentials = service_account.Credentials.from_service_account_file(auth_file)
            auth_data = self.read_auth_file(auth_file)
            project_id = auth_data.get('project_id')
            account_email = auth_data.get('client_email')
            return credentials, project_id, account_email
        else:
            raise GCPDriverError(f"file auth selected: can not find application_default_credentials.json")

    @staticmethod
    def read_auth_file(auth_file: str):
        file_handle = open(auth_file, 'r')
        auth_data = json.load(file_handle)
        file_handle.close()
        return auth_data

    @retry()
    def zones(self) -> list:
        try:
            request = self.gcp_client.zones().list(project=self.gcp_project)
            while request is not None:
                response = request.execute()
                for zone in response['items']:
                    if not zone['name'].startswith(self.gcp_region):
                        continue
                    self.gcp_zone_list.append(zone['name'])
                request = self.gcp_client.zones().list_next(previous_request=request, previous_response=response)
        except Exception as err:
            raise GCPDriverTransientError(f"error getting zones: {err}")

        self.gcp_zone_list = sorted(set(self.gcp_zone_list))

        if len(self.gcp_zone_list) == 0:
            raise GCPDriverError("can not get GCP availability zones")

        self.gcp_zone = self.gcp_zone_list[0]
        return self.gcp_zone_list

    @property
    def region(self):
        return self.gcp_region

    @property
    def account_file(self):
        return self.gcp_account_file

    @property
    def project(self):
        return self.gcp_project

    @property
    def service_account_email(self):
        return self._service_account_email

    @property
    def login_account_email(self):
        return self._user_account_email

    @property
    def account_email(self):
        return self._service_account_email if self._service_account_email else self._user_account_email

    @staticmethod
    def process_labels(struct: dict) -> dict:
        block = {}
        if 'labels' in struct:
            for tag in struct['labels']:
                block.update({tag.lower() + '_tag': struct['labels'][tag]})
        block = dict(sorted(block.items()))
        return block

    def wait_for_global_operation(self, operation):
        while True:
            result = self.gcp_client.globalOperations().get(
                project=self.gcp_project,
                operation=operation).execute()

            if result['status'] == 'DONE':
                if 'error' in result:
                    raise GCPDriverError(result['error'])
                return result

            time.sleep(1)

    def wait_for_regional_operation(self, operation):
        while True:
            result = self.gcp_client.regionOperations().get(
                project=self.gcp_project,
                region=self.gcp_region,
                operation=operation).execute()

            if result['status'] == 'DONE':
                if 'error' in result:
                    raise GCPDriverError(result['error'])
                return result

            time.sleep(1)

    def wait_for_zone_operation(self, operation, zone):
        while True:
            result = self.gcp_client.zoneOperations().get(
                project=self.gcp_project,
                zone=zone,
                operation=operation).execute()

            if result['status'] == 'DONE':
                if 'error' in result:
                    raise GCPDriverError(result['error'])
                return result

            time.sleep(1)



##
##

import os
from typing import Optional, Union, List
from enum import Enum
from pathlib import Path
import attr
import argparse
import yaml
import itertools
import couchformation.constants as C


def get_base_dir():
    if 'COUCH_FORMATION_CONFIG_DIR' in os.environ:
        return os.environ['COUCH_FORMATION_CONFIG_DIR']
    else:
        return C.STATE_DIRECTORY


def get_log_dir():
    if 'COUCH_FORMATION_LOG_DIR' in os.environ:
        return os.environ['COUCH_FORMATION_LOG_DIR']
    else:
        return C.LOG_DIRECTORY


def get_root_dir():
    return C.ROOT_DIRECTORY


def get_resource_dir(name: str, tag: str):
    return os.path.join(get_base_dir(), name, tag)


def get_project_dir(name: str):
    return os.path.join(get_base_dir(), name)


def get_state_file(project: str, name: str):
    return os.path.join(get_project_dir(project), name, C.STATE)


def get_state_dir(project: str, name: str):
    return os.path.join(get_project_dir(project), name)


def str_to_int(value: Union[str, int]) -> int:
    return int(value)


class AuthMode(Enum):
    default = 0
    sso = 1
    file = 2


class PathMode(Enum):
    resource = 0
    common = 1


class ProvisionMode(Enum):
    public = 0
    private = 1


class State(Enum):
    IDLE = 0
    DEPLOYING = 1
    DEPLOYED = 2
    DESTROYING = 3
    FAILED = 4
    ERROR = 5


@attr.s
class Parameters:
    project: Optional[str] = attr.ib(default=None)
    cloud: Optional[str] = attr.ib(default=None)
    name: Optional[str] = attr.ib(default=None)
    model: Optional[str] = attr.ib(default=None)
    region: Optional[str] = attr.ib(default=None)
    ssh_key: Optional[str] = attr.ib(default=None)
    os_id: Optional[str] = attr.ib(default=None)
    os_version: Optional[str] = attr.ib(default=None)
    auth_mode: Optional[str] = attr.ib(default=None)
    profile: Optional[str] = attr.ib(default=None)
    base_dir: Optional[str] = attr.ib(default=None)
    private_ip: Optional[bool] = attr.ib(default=None)
    path_mode: Optional[PathMode] = attr.ib(default=None)
    machine_type: Optional[str] = attr.ib(default=None)
    quantity: Optional[int] = attr.ib(default=None)
    services: Optional[str] = attr.ib(default=None)
    volume_iops: Optional[str] = attr.ib(default=None)
    volume_size: Optional[str] = attr.ib(default=None)
    volume_type: Optional[str] = attr.ib(default=None)
    volume_tier: Optional[str] = attr.ib(default=None)
    root_size: Optional[str] = attr.ib(default=None)
    connect_svc: Optional[str] = attr.ib(default=None)
    connect_ip: Optional[str] = attr.ib(default=None)

    @classmethod
    def create(cls, args):
        c = cls()
        c.initialize_args(args)
        return c

    def initialize_args(self, args):
        parser = argparse.ArgumentParser(add_help=False)
        for attribute in self.__annotations__:
            parser.add_argument(f"--{attribute}", action='store')
        parameters, remainder = parser.parse_known_args(args)
        self.from_namespace(parameters)

    def from_namespace(self, namespace: argparse.Namespace):
        args = vars(namespace)
        for attribute in self.__annotations__:
            if args.get(attribute):
                setattr(self, attribute, args.get(attribute))

    def from_dict(self, options: dict):
        for attribute in self.__annotations__:
            if options.get(attribute):
                setattr(self, attribute, options.get(attribute))

    @property
    def project_dir(self):
        return get_project_dir(self.project)

    @property
    def as_dict(self):
        return {k: self.__dict__[k] for k in self.__dict__ if self.__dict__[k] is not None}


@attr.s
class BaseConfig:
    project: Optional[str] = attr.ib(default="resources")
    ssh_key: Optional[str] = attr.ib(default=os.path.join(Path.home(), '.ssh', 'couch-formation-key.pem'))
    base_dir: Optional[str] = attr.ib(default=get_base_dir())
    private_ip: Optional[bool] = attr.ib(default=False)

    @classmethod
    def create(cls, data: Union[list, dict]):
        if type(data) is list:
            c = cls()
            c.initialize_args(data)
        else:
            c = cls()
            c.initialize_dict(data)
        return c

    def initialize_args(self, args):
        parser = argparse.ArgumentParser(add_help=False)
        for attribute in self.__annotations__:
            parser.add_argument(f"--{attribute}", action='store')
        parameters, remainder = parser.parse_known_args(args)
        self.from_namespace(parameters)

    def initialize_dict(self, options):
        self.from_dict(options)

    @property
    def project_dir(self):
        return get_project_dir(self.project)

    def from_namespace(self, namespace: argparse.Namespace):
        args = vars(namespace)
        for attribute in self.__annotations__:
            if args.get(attribute):
                setattr(self, attribute, args.get(attribute))

    def from_dict(self, options: dict):
        for attribute in self.__annotations__:
            if options.get(attribute):
                setattr(self, attribute, options.get(attribute))

    @property
    def as_dict(self):
        return self.__dict__


@attr.s
class NodeConfig:
    machine_type: Optional[str] = attr.ib(default=None)
    quantity: Optional[int] = attr.ib(default=1, converter=str_to_int)
    services: Optional[str] = attr.ib(default="default")
    volume_iops: Optional[str] = attr.ib(default="3000")
    volume_size: Optional[str] = attr.ib(default="256")
    volume_type: Optional[str] = attr.ib(default=None)
    root_size: Optional[str] = attr.ib(default="256")

    @classmethod
    def create(cls, data: Union[list, dict]):
        if type(data) is list:
            c = cls()
            c.initialize_args(data)
        else:
            c = cls()
            c.initialize_dict(data)
        return c

    def initialize_args(self, args):
        parser = argparse.ArgumentParser(add_help=False)
        for attribute in self.__annotations__:
            parser.add_argument(f"--{attribute}", action='store')
        parameters, undefined = parser.parse_known_args(args)
        self.from_namespace(parameters)

    def initialize_dict(self, options):
        self.from_dict(options)

    def from_namespace(self, namespace: argparse.Namespace):
        args = vars(namespace)
        for attribute in self.__annotations__:
            if args.get(attribute):
                setattr(self, attribute, args.get(attribute))

    def from_dict(self, options: dict):
        for attribute in self.__annotations__:
            if options.get(attribute):
                setattr(self, attribute, options.get(attribute))

    @property
    def as_dict(self):
        return self.__dict__


@attr.s
class DeploymentConfig:
    core: Optional[BaseConfig] = attr.ib(default=BaseConfig())
    config: Optional[List[NodeConfig]] = attr.ib(default=[])

    def add_config(self, index: Union[str, int], config: NodeConfig):
        index = int(index)
        config.group = str(index)
        if index > len(self.config) + 1:
            raise ValueError(f"config index {index} out of range")
        elif index == len(self.config) + 1:
            self.config.append(config)
        else:
            self.config[index - 1] = config

    @classmethod
    def new(cls, core: BaseConfig):
        return cls(
            core,
            []
        )

    def reset(self, args):
        self.config.clear()
        self.core = BaseConfig().create(args)

    @property
    def length(self):
        return len(self.config)

    @property
    def as_dict(self):
        return self.__dict__


@attr.s
class NodeEntry:
    name: Optional[str] = attr.ib(default=None)
    username: Optional[str] = attr.ib(default=None)
    private_ip: Optional[str] = attr.ib(default=None)
    public_ip: Optional[str] = attr.ib(default=None)
    use_private_ip: Optional[bool] = attr.ib(default=False)
    availability_zone: Optional[str] = attr.ib(default=None)
    services: Optional[str] = attr.ib(default=None)
    connect_svc: Optional[str] = attr.ib(default=None)
    connect_ip: Optional[str] = attr.ib(default=None)

    @classmethod
    def create(cls,
               name: str,
               username: str,
               private_ip: str,
               public_ip: str = None,
               use_private_ip: bool = False,
               zone: str = None,
               services: str = "default",
               connect_svc: str = None,
               connect_ip: str = None
               ):
        return cls(
            name,
            username,
            private_ip,
            public_ip,
            use_private_ip,
            zone,
            services,
            connect_svc,
            connect_ip
        )


@attr.s
class NodeList:
    username: Optional[str] = attr.ib(default=None)
    ssh_key: Optional[str] = attr.ib(default=None)
    nodes: Optional[List[NodeEntry]] = attr.ib(default=[])
    working_dir: Optional[str] = attr.ib(default=None)
    provision_ip: Optional[ProvisionMode] = attr.ib(default=ProvisionMode.public)

    @classmethod
    def create(cls, username: str, ssh_key: str, working_dir: str = None, use_private_ip: bool = False):
        return cls(
            username,
            ssh_key,
            [],
            working_dir,
            ProvisionMode(use_private_ip)
        )

    def add(self, name: str, private_ip: str, public_ip: str = None, zone: str = None, services: str = "default", connect_svc: str = None, connect_ip: str = None):
        self.nodes.append(
            NodeEntry.create(
                name,
                self.username,
                private_ip,
                public_ip,
                bool(self.provision_ip.value),
                zone,
                services,
                connect_svc,
                connect_ip
            )
        )

    @property
    def node_list(self) -> List[NodeEntry]:
        return self.nodes

    def list_public_ip(self):
        address_list = []
        for entry in self.nodes:
            address_list.append(entry.public_ip)
        return address_list

    def list_private_ip(self):
        address_list = []
        for entry in self.nodes:
            address_list.append(entry.private_ip)
        return address_list

    def provision_list(self):
        if self.provision_ip == ProvisionMode.public:
            return self.list_public_ip()
        else:
            return self.list_private_ip()

    def ip_csv_list(self):
        return ','.join(self.list_private_ip())


@attr.s
class PortSettings:
    build: Optional[str] = attr.ib(default=None)
    tcp_ports: Optional[List[int]] = attr.ib(default=[])
    udp_ports: Optional[List[int]] = attr.ib(default=[])

    @classmethod
    def create(cls, name: str, port_config: List[str]):
        tcp_list = []
        udp_list = []
        tcp_ports = []
        udp_ports = []
        for spec in port_config:
            spec = str(spec)
            port_spec = spec.split('/')
            port_nums = port_spec[0]
            if len(port_spec) > 1:
                port_proto = port_spec[1]
            else:
                port_proto = 'tcp'
            if port_proto == 'tcp':
                tcp_list.append(port_nums)
            else:
                udp_list.append(port_nums)
        for ports in tcp_list:
            port_split = ports.split('-')
            if len(port_split) > 1:
                tcp_ports.extend(list(range(int(port_split[0]), int(port_split[1])+1)))
            else:
                tcp_ports.append(int(port_split[0]))
        for ports in udp_list:
            port_split = ports.split('-')
            if len(port_split) > 1:
                udp_ports.extend(list(range(int(port_split[0]), int(port_split[1])+1)))
            else:
                udp_ports.append(int(port_split[0]))
        tcp_ports.sort()
        udp_ports.sort()

        return cls(
            name,
            tcp_ports,
            udp_ports
        )

    @property
    def has_tcp_ports(self):
        return len(self.tcp_ports) > 0

    @property
    def has_udp_ports(self):
        return len(self.udp_ports) > 0

    @staticmethod
    def ranges(ports: List[int]):
        for a, b in itertools.groupby(enumerate(ports), lambda pair: pair[1] - pair[0]):
            b = list(b)
            yield b[0][1], b[-1][1]

    def tcp_as_tuple(self):
        for begin, end in self.ranges(self.tcp_ports):
            yield begin, end

    def udp_as_tuple(self):
        for begin, end in self.ranges(self.udp_ports):
            yield begin, end

    def tcp_as_ranges(self):
        for begin, end in self.ranges(self.tcp_ports):
            if begin < end:
                yield f"{begin}-{end}"
            else:
                yield str(begin)

    def udp_as_ranges(self):
        for begin, end in self.ranges(self.udp_ports):
            if begin < end:
                yield f"{begin}-{end}"
            else:
                yield str(begin)


@attr.s
class PortSettingSet:
    config: Optional[List[PortSettings]] = attr.ib(default=[])

    @classmethod
    def create(cls):
        profile = []
        with open(C.NODE_PORT_PROFILE, "r") as f:
            try:
                for build, settings in yaml.safe_load(f).items():
                    port_settings = PortSettings.create(build, settings.get('ports', []))
                    profile.append(port_settings)
                return cls(
                    profile
                )
            except yaml.YAMLError as err:
                raise RuntimeError(f"Can not open port config file {C.NODE_PORT_PROFILE}: {err}")

    def get(self, name: str):
        return next((o for o in self.config if o.build == name), None)

    def items(self):
        return self.config



##
##

import logging
import warnings
import argparse
import sys
import os
import re
import signal
import inspect
import traceback
import datetime
import psutil
import logging.handlers
from datetime import datetime, timezone
from couchformation.config import get_log_dir
import couchformation.constants as C
from couchformation.util import FileManager
from couchformation.exception import FatalError

warnings.filterwarnings("ignore")
logger = logging.getLogger()


def break_signal_handler(signum, frame):
    signal_name = signal.Signals(signum).name
    (filename, line, function, lines, index) = inspect.getframeinfo(frame)
    logger.info(f"received signal {signal_name} in {filename} {function} at line {line}")
    tb = ''.join(traceback.format_stack(frame))
    logger.info(tb)
    print("")
    print("Break received, aborting.")
    sys.exit(1)


class CloudMgrError(FatalError):
    pass


class CustomDisplayFormatter(logging.Formatter):
    FORMATS = {
        logging.DEBUG: f"[{C.GREY_COLOR}{C.FORMAT_LEVEL}{C.SCREEN_RESET}] {C.FORMAT_MESSAGE}",
        logging.INFO: f"[{C.GREEN_COLOR}{C.FORMAT_LEVEL}{C.SCREEN_RESET}] {C.FORMAT_MESSAGE}",
        logging.WARNING: f"[{C.YELLOW_COLOR}{C.FORMAT_LEVEL}{C.SCREEN_RESET}] {C.FORMAT_MESSAGE}",
        logging.ERROR: f"[{C.RED_COLOR}{C.FORMAT_LEVEL}{C.SCREEN_RESET}] {C.FORMAT_MESSAGE}",
        logging.CRITICAL: f"[{C.BOLD_RED_COLOR}{C.FORMAT_LEVEL}{C.SCREEN_RESET}] {C.FORMAT_MESSAGE}"
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


class CustomMinimalDisplayFormatter(logging.Formatter):
    FORMATS = {
        logging.DEBUG: f"{C.FORMAT_MESSAGE}",
        logging.INFO: f"{C.FORMAT_MESSAGE}",
        logging.WARNING: f"{C.FORMAT_MESSAGE}",
        logging.ERROR: f"{C.FORMAT_MESSAGE}",
        logging.CRITICAL: f"{C.FORMAT_MESSAGE}"
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


class CustomLogFormatter(logging.Formatter):
    FORMATS = {
        logging.DEBUG: f"{C.FORMAT_TIMESTAMP} [{C.FORMAT_LEVEL}] ({C.FORMAT_THREAD}) {C.FORMAT_MESSAGE}",
        logging.INFO: f"{C.FORMAT_TIMESTAMP} [{C.FORMAT_LEVEL}] ({C.FORMAT_THREAD}) {C.FORMAT_MESSAGE}",
        logging.WARNING: f"{C.FORMAT_TIMESTAMP} [{C.FORMAT_LEVEL}] ({C.FORMAT_THREAD}) {C.FORMAT_MESSAGE}",
        logging.ERROR: f"{C.FORMAT_TIMESTAMP} [{C.FORMAT_LEVEL}] ({C.FORMAT_THREAD}) {C.FORMAT_MESSAGE}",
        logging.CRITICAL: f"{C.FORMAT_TIMESTAMP} [{C.FORMAT_LEVEL}] ({C.FORMAT_THREAD}) {C.FORMAT_MESSAGE}"
    }

    @staticmethod
    def _filter(s):
        return re.sub(r'(\".*password\"\s*:\s*\").+?(\")', r'\1********\2', s)

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        if logging.DEBUG >= logging.root.level:
            log_fmt += C.FORMAT_EXTRA
        formatter = logging.Formatter(log_fmt)
        original = formatter.format(record)
        return self._filter(original)


class StreamOutputLogger(object):
    def __init__(self, _logger, _level, _file=None):
        self.logger = _logger
        self.level = _level
        if not _file:
            self.file = sys.stdout
        else:
            self.file = _file
        self.buffer = ''

    def write(self, buf):
        for line in buf.rstrip().splitlines():
            self.logger.log(self.level, line.rstrip())

    def __getattr__(self, name):
        return getattr(self.file, name)

    def flush(self):
        pass


class CLI(object):

    def __init__(self, args):
        self.log_dir = get_log_dir()
        signal.signal(signal.SIGINT, break_signal_handler)
        default_debug_file = os.path.join(self.log_dir, "formation.log")
        debug_file = os.environ.get("COUCH_FORMATION_DEBUG_FILE", default_debug_file)
        self.args = args
        self.parser = None
        self.options = None
        self.remainder = None

        if os.name == 'nt':
            from ctypes import windll
            k = windll.kernel32
            k.SetConsoleMode(k.GetStdHandle(-11), 7)

        if self.args is None:
            self.args = sys.argv[1:]

        self.init_parser()

        try:
            FileManager().make_dir(self.log_dir)
        except Exception as err:
            raise CloudMgrError(f"can not create log dir: {err}")

        screen_handler = logging.StreamHandler()
        screen_handler.setFormatter(CustomDisplayFormatter())
        screen_handler.setLevel(logging.INFO)
        logger.addHandler(screen_handler)

        min_log = logging.getLogger('minimum_output')
        min_screen_handler = logging.StreamHandler()
        min_screen_handler.setFormatter(CustomMinimalDisplayFormatter())
        min_screen_handler.setLevel(logging.INFO)
        min_log.addHandler(min_screen_handler)
        min_log.propagate = False

        file_handler = logging.handlers.RotatingFileHandler(debug_file, maxBytes=10485760, backupCount=5)
        file_handler.setFormatter(CustomLogFormatter())
        logger.addHandler(file_handler)

        logger.setLevel(logging.DEBUG)

        process = psutil.Process(os.getpid())
        logger.debug(f"---- CLI Start: {self.get_timestamp()} PID: {process.pid}----")
        logger.debug(f"---- CLI: {process.name()} {' '.join(sys.argv)} ----")

        self.process_args()

    @staticmethod
    def get_timestamp():
        return datetime.now(timezone.utc).strftime("%b %d %H:%M:%S")

    def init_parser(self):
        self.parser = argparse.ArgumentParser(add_help=False)
        self.parser.add_argument('-d', '--debug', action='store_true', help="Debug output")
        self.parser.add_argument('-v', '--verbose', action='store_true', help="Verbose output")

    def local_args(self):
        pass

    def process_args(self):
        self.local_args()
        self.options, self.remainder = self.parser.parse_known_args(self.args)
        if self.options.debug:
            logger.setLevel(logging.DEBUG)



##
##

import logging
import os
import couchformation.constants as C
import couchformation.kvdb as kvdb
from couchformation.config import get_root_dir
from couchformation.exception import FatalError
from couchformation.util import FileManager
from couchformation.kvdb import KeyValueStore

logger = logging.getLogger('couchformation.config.manager')
logger.addHandler(logging.NullHandler())


class ConfigError(FatalError):
    pass


PARAMETERS = {
    'capella.token': {
        'type': 'string',
        'mutable': True
    },
    'capella.user': {
        'type': 'string',
        'mutable': True
    },
    'capella.user.id': {
        'type': 'string',
        'mutable': True
    },
    'capella.project': {
        'type': 'string',
        'mutable': True
    }
}


class ConfigurationManager(object):

    def __init__(self):
        self.filename = C.CONFIG_FILE

        try:
            if not os.path.exists(get_root_dir()):
                FileManager().make_dir(get_root_dir())
        except Exception as err:
            raise ConfigError(f"can not create root dir: {err}")

    @staticmethod
    def key_split(key: str):
        vector = key.split('.', 1)
        if len(vector) != 2:
            raise ConfigError(f"Malformed key {key}")

        table_name = vector[0]
        value_name = vector[1]

        return table_name, value_name

    @staticmethod
    def strtobool(val):
        val = val.lower()
        if val in ('y', 'yes', 't', 'true', 'on', '1'):
            return 1
        elif val in ('n', 'no', 'f', 'false', 'off', '0'):
            return 0
        else:
            raise ValueError(f"invalid truth value {val}")

    def convert(self, key, value):
        key_settings = PARAMETERS.get(key)
        if key_settings is None:
            raise ConfigError(f"Configuration key {key} not supported")

        try:
            if key_settings['type'] == 'string':
                return str(value)
            elif key_settings['type'] == 'integer':
                return int(value)
            elif key_settings['type'] == 'boolean':
                return bool(self.strtobool(value))
            elif key_settings['type'] == 'decimal':
                return float(value)
        except ValueError as e:
            raise ConfigError(f"Invalid value {value} for key {key} of type {value}: {e}")

    @staticmethod
    def mutable(key: str):
        key_settings = PARAMETERS.get(key)
        if key_settings is None:
            raise ConfigError(f"Configuration key {key} not supported")
        return key_settings['mutable']

    def get(self, key: str):
        if key not in PARAMETERS:
            raise ConfigError(f"Configuration parameter {key} not supported")

        table_name, value_name = self.key_split(key)

        table = KeyValueStore(self.filename, table_name)
        if table.get(value_name):
            return self.convert(key, table.get(value_name))
        else:
            return None

    def set(self, key: str, value: str):
        if key not in PARAMETERS:
            raise ConfigError(f"Configuration parameter {key} not supported")
        if not self.mutable(key):
            raise ConfigError(f"Configuration key {key} is not mutable")

        table_name, value_name = self.key_split(key)

        table = KeyValueStore(self.filename, table_name)
        table[value_name] = self.convert(key, value)

    def delete(self, key: str):
        if key not in PARAMETERS:
            raise ConfigError(f"Configuration parameter {key} not supported")
        if not self.mutable(key):
            raise ConfigError(f"Configuration key {key} is not mutable")

        table_name, value_name = self.key_split(key)

        table = KeyValueStore(self.filename, table_name)
        if table.get(value_name):
            del table[value_name]

    def reset(self):
        for table in kvdb.documents(self.filename):
            table.clear()

    def list(self):
        response = {}
        for key in PARAMETERS.keys():
            table_name, value_name = self.key_split(key)
            table = KeyValueStore(self.filename, table_name)
            if table.get(value_name):
                response[key] = self.convert(key, table.get(value_name))
        return response



##
##

import io
import os
import logging
import docker
import jinja2
import tarfile
from jinja2.meta import find_undeclared_variables
from docker.errors import APIError
from docker.models.containers import Container as ContainerClass
from typing import Union, List, Set
from couchformation.docker.driver.base import CloudBase, DockerDriverError
from couchformation.docker.util import ContainerProfile, ContainerSpec
from couchformation.util import FileManager
from couchformation.config import get_project_dir

logger = logging.getLogger('couchformation.docker.driver.container')
logger.addHandler(logging.NullHandler())
logging.getLogger("docker").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)


class Container(CloudBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def list(self, name: str = None) -> Union[List[dict], None]:
        container_list = []

        try:
            containers = self.client.containers.list()
        except Exception as err:
            raise DockerDriverError(f"error getting container list: {err}")

        for container in containers:
            if name and container.name != name:
                continue
            container_block = {'name': container.name,
                               'short_id': container.short_id,
                               'status': container.status,
                               'ports': self.get_container_ports(container.name),
                               'ip_address': self.get_container_ip(container.name),
                               'id': container.id}
            container_list.append(container_block)

        if len(container_list) == 0:
            return None
        else:
            return container_list

    @staticmethod
    def expand_ranges(text):
        range_list = []
        for i in text.split(','):
            if '-' not in i:
                range_list.append(int(i))
            else:
                l, h = map(int, i.split('-'))
                range_list += range(l, h + 1)
        return range_list

    @staticmethod
    def create_port_dict(port_list: List[int]):
        result = {}
        for port in port_list:
            key = f"{port}/tcp"
            result.update({key: port})
        return result

    def run(self,
            image: str,
            name: str,
            volume_mount: Union[str, None] = None,
            volume_size: int = 2048,
            dir_mount: Union[str, None] = None,
            platform: Union[str, None] = None,
            ports: Union[str, None] = None,
            network: Union[str, None] = None,
            command: Union[str, list, None] = None) -> ContainerClass:
        if self.map(image):
            image = self.map(image)
        cp = ContainerProfile()
        ip = cp.get(image)
        volume_map = None

        if not platform:
            platform = self.arch
        if not ports:
            if ip.ports:
                ports = ip.ports
            else:
                ports = "80,443"
        if not volume_mount and ip.volume:
            volume_mount = ip.volume.directory
            if ip.volume.size:
                volume_size = ip.volume.size
        if not dir_mount and ip.volume and ip.volume.run:
            host_path = os.path.join(get_project_dir(self.project), self.name, 'mounts', name)
            dir_mount = host_path
            if ip.volume.run:
                run_cmd = self.process_command_vars(ip)
            else:
                run_cmd = None
            FileManager().dir_populate(str(host_path), run_cmd)
        if not command and ip.volume and ip.volume.command:
            command = ip.volume.command

        port_struct = self.create_port_dict(self.expand_ranges(ports))

        try:
            if volume_mount and not dir_mount:
                volume = self.client.volumes.create(name=f"{name}-vol", driver="local", driver_opts={"type": "tmpfs", "device": "tmpfs", "o": f"size={volume_size}m"})
                volume_map = [f"{volume.name}:{volume_mount}"]
            elif dir_mount:
                if not volume_mount:
                    volume_mount = dir_mount
                volume_map = [f"{dir_mount}:{volume_mount}"]
            container_id = self.client.containers.run(image,
                                                      tty=True,
                                                      detach=True,
                                                      platform=platform,
                                                      name=name,
                                                      ports=port_struct,
                                                      network=network,
                                                      volumes=volume_map,
                                                      command=command
                                                      )
        except docker.errors.APIError as e:
            if e.status_code == 409:
                container_id = self.client.containers.get(name)
            else:
                raise

        logger.debug("Container started")
        return container_id

    @staticmethod
    def map(build: str):
        try:
            return ContainerProfile().by_build(build).name
        except AttributeError:
            return None

    def get_container(self, container_id: str):
        return self.get_container_id(container_id)

    @staticmethod
    def get_container_id(name: str) -> Union[ContainerClass, None]:
        client = docker.from_env()
        try:
            return client.containers.get(name)
        except docker.errors.NotFound:
            return None

    def get_container_ip(self, name: str):
        container_id = self.get_container_id(name)
        try:
            ip_address = container_id.attrs['NetworkSettings']['IPAddress']
            if len(ip_address) == 0:
                networks = container_id.attrs.get('NetworkSettings', {}).get('Networks', {})
                for network, settings in networks.items():
                    ip_address = settings.get('IPAddress')
                    if len(ip_address) > 0:
                        break
            return ip_address
        except KeyError:
            return None

    def get_container_ports(self, name: str):
        container_id = self.get_container_id(name)
        ports = list(p.split('/')[0] for p in container_id.attrs.get('Config', {}).get('ExposedPorts', {}).keys())
        return ports

    @staticmethod
    def get_file_parameters(input_text) -> Set[str]:
        env = jinja2.Environment(undefined=jinja2.DebugUndefined)
        template = env.from_string(input_text)
        rendered = template.render()
        ast = env.parse(rendered)
        return find_undeclared_variables(ast)

    def process_command_vars(self, profile: ContainerSpec) -> str:
        tags = []
        variables = self.get_file_parameters(profile.volume.run)
        for variable in variables:
            if variable.startswith('port'):
                ip_list = []
                try:
                    check_port = variable.split('_')[1]
                except IndexError:
                    check_port = "0"
                for container in self.list():
                    if check_port in container.get('ports'):
                        ip_list.append(container.get('ip_address'))
                ip_list_str = ','.join(ip_list)
                tags.append((variable, ip_list_str))
            elif variable == 'dir_name' and profile.volume.directory:
                tags.append((variable, profile.volume.directory))
        parameters = dict((a, b) for a, b in tags)
        raw_template = jinja2.Template(profile.volume.run)
        return raw_template.render(parameters)

    def run_in_container(self, name: str, command: Union[str, List[str]], directory: Union[str, None] = None, root: bool = True):
        buffer = io.BytesIO()
        cmd_prefix = ['sh', '-c']
        container_id = self.get_container_id(name)
        if type(command) is str:
            command = [command]
        cmd_prefix.extend(command)
        if root:
            exit_code, output = container_id.exec_run(cmd_prefix, workdir=directory, user="root")
        else:
            exit_code, output = container_id.exec_run(cmd_prefix, workdir=directory)
        buffer.write(output)
        buffer.seek(0)
        return exit_code, buffer

    def get_home_path(self, name: str) -> Union[str, None]:
        exit_code, output = self.run_in_container(name, "printf %s ~")
        if exit_code != 0:
            return None
        else:
            return output.read().decode('utf-8')

    def copy_file_to_container(self, name: str, src: str, dst: str):
        container_id = self.get_container_id(name)
        stream = io.BytesIO()
        with tarfile.open(fileobj=stream, mode='w|') as tar, open(src, 'rb') as file:
            info = tar.gettarinfo(fileobj=file)
            info.name = os.path.basename(src)
            tar.addfile(info, file)
        container_id.put_archive(dst, stream.getvalue())

    def copy_io_to_container(self, name: str, fl: io, dst: str):
        fl.seek(0)
        container_id = self.get_container_id(name)
        home_path = self.get_home_path(name)
        if not home_path:
            home_path = "/root"
        stream = io.BytesIO()
        with tarfile.open(fileobj=stream, mode='w|') as tar:
            info = tarfile.TarInfo(name=dst)
            info.size = fl.getbuffer().nbytes
            tar.addfile(info, fl)
        container_id.put_archive(home_path, stream.getvalue())

    def terminate(self, name: str) -> None:
        container_id = self.get_container_id(name)
        if not container_id:
            return
        client = docker.from_env()
        container_id.stop()
        container_id.remove()
        try:
            volume = client.volumes.get(f"{name}-vol")
            volume.remove()
        except docker.errors.NotFound:
            pass



##
##

import logging
import re
from typing import Union, List
from azure.core.exceptions import ResourceNotFoundError
from azure.mgmt.network.models import VirtualNetwork
from couchformation.azure.driver.base import CloudBase, AzureDriverError, EmptyResultSet

logger = logging.getLogger('couchformation.azure.driver.network')
logger.addHandler(logging.NullHandler())
logging.getLogger("azure").setLevel(logging.ERROR)
logging.getLogger("urllib3").setLevel(logging.ERROR)


class Network(CloudBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger = logging.getLogger(self.__class__.__name__)

    def list(self, resource_group: Union[str, None] = None, filter_keys_exist: Union[List[str], None] = None) -> List[dict]:
        if not resource_group:
            if not self.azure_resource_group:
                return []
            resource_group = self.azure_resource_group
        vnet_list = []

        try:
            vnetworks = self.network_client.virtual_networks.list(resource_group)
        except Exception as err:
            raise AzureDriverError(f"error getting vnet: {err}")

        for group in list(vnetworks):
            if group.location != self.azure_location:
                continue
            network_block = {'cidr': group.address_space.address_prefixes,
                             'name': group.name,
                             'subnets': [s.name for s in group.subnets],
                             'id': group.id}
            network_block.update(self.process_tags(group.tags))
            if filter_keys_exist:
                if not all(key in network_block for key in filter_keys_exist):
                    continue
            vnet_list.append(network_block)

        if len(vnet_list) == 0:
            raise EmptyResultSet(f"no suitable virtual network in location {self.azure_location}")

        return vnet_list

    @property
    def cidr_list(self):
        try:
            for item in self.list():
                for net in item['cidr']:
                    yield net
        except EmptyResultSet:
            return iter(())

    def create(self, name: str, cidr: str, resource_group: str):
        parameters = {
            'location': self.azure_location,
            'address_space': {
                'address_prefixes': [cidr]
            }
        }

        net_info = self.details(name, resource_group)
        if net_info:
            return net_info

        try:
            request = self.network_client.virtual_networks.begin_create_or_update(resource_group, name, parameters)
            request.wait()
            return request.result()
        except Exception as err:
            raise AzureDriverError(f"error creating network: {err}")

    def delete(self, network: str, resource_group: str) -> None:
        try:
            request = self.network_client.virtual_networks.begin_delete(resource_group, network)
            request.wait()
        except ResourceNotFoundError:
            return None
        except Exception as err:
            raise AzureDriverError(f"error getting vnet: {err}")

    def details(self, network: str, resource_group: str) -> Union[VirtualNetwork, None]:
        try:
            info = self.network_client.virtual_networks.get(resource_group, network)
            return info
        except ResourceNotFoundError:
            return None
        except Exception as err:
            raise AzureDriverError(f"error getting vnet: {err}")

    def create_pub_ip(self, name: str, resource_group: str):
        parameters = {
            'location': self.azure_location,
            'public_ip_allocation_method': 'Static',
            'sku': {
                'name': 'Standard'
            }
        }

        try:
            request = self.network_client.public_ip_addresses.begin_create_or_update(resource_group, name, parameters)
            request.wait()
            return request.result()
        except Exception as err:
            raise AzureDriverError(f"can not create public IP: {err}")

    def create_nic(self, name: str, subnet_id: str, zone: str, pub_ip_id: str, resource_group: str):
        parameters = {
            'location': self.azure_location,
            'ip_configurations': [
                {
                    'name': name,
                    'subnet': {
                        'id': subnet_id,
                    },
                    'private_ip_allocation_method': 'Dynamic',
                    'zones': [zone],
                    'public_ip_address': {
                        'id': pub_ip_id
                    }
                }
            ]
        }

        try:
            request = self.network_client.network_interfaces.begin_create_or_update(resource_group, name, parameters)
            request.wait()
            return request.result()
        except Exception as err:
            raise AzureDriverError(f"error creating nic: {err}")

    def describe_nic(self, name: str, resource_group: str):
        try:
            nic = self.network_client.network_interfaces.get(resource_group, name)
            return nic
        except ResourceNotFoundError:
            return None
        except Exception as err:
            raise AzureDriverError(f"error getting nic details: {err}")

    def describe_pub_ip(self, name: str, resource_group: str):
        try:
            ip = self.network_client.public_ip_addresses.get(resource_group, name)
            return ip
        except ResourceNotFoundError:
            return None
        except Exception as err:
            raise AzureDriverError(f"error getting public IP: {err}")

    def delete_pub_ip(self, name: str, resource_group: str):
        try:
            request = self.network_client.public_ip_addresses.begin_delete(resource_group, name)
            request.wait()
        except ResourceNotFoundError:
            return None
        except Exception as err:
            raise AzureDriverError(f"error deleting public IP: {err}")

    def delete_nic(self, name: str, resource_group: str) -> None:
        try:
            request = self.network_client.network_interfaces.begin_delete(resource_group, name)
            request.wait()
        except ResourceNotFoundError:
            return None
        except Exception as err:
            raise AzureDriverError(f"error deleting nic: {err}")


class Subnet(CloudBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger = logging.getLogger(self.__class__.__name__)

    def list(self, vnet: str, resource_group: Union[str, None] = None) -> List[dict]:
        if not resource_group:
            resource_group = self.azure_resource_group
        subnet_list = []

        try:
            subnets = self.network_client.subnets.list(resource_group, vnet)
        except Exception as err:
            raise AzureDriverError(f"error getting subnets: {err}")

        for group in list(subnets):
            subnet_block = {'cidr': group.address_prefix,
                            'name': group.name,
                            'routes': group.route_table.routes if group.route_table else None,
                            'nsg': group.network_security_group.id.rsplit('/', 1)[-1] if group.network_security_group else None,
                            'id': group.id}
            subnet_list.append(subnet_block)

        if len(subnet_list) == 0:
            raise EmptyResultSet(f"no subnets in vnet {vnet}")

        return subnet_list

    def create(self, name: str, network: str, cidr: str, nsg_id: str, resource_group: str):
        parameters = {
            'address_prefix': cidr,
            'network_security_group': {
                'id': nsg_id
            }
        }
        try:
            request = self.network_client.subnets.begin_create_or_update(resource_group, network, name, parameters)
            request.wait()
            return request.result()
        except Exception as err:
            raise AzureDriverError(f"error creating subnet: {err}")

    def delete(self, network: str, subnet: str, resource_group: str) -> None:
        try:
            subnet_info = self.details(network, subnet, resource_group)
            if subnet_info:
                request = self.network_client.subnets.begin_delete(resource_group, network, subnet)
                request.wait()
        except ResourceNotFoundError:
            return None
        except Exception as err:
            raise AzureDriverError(f"error deleting subnet: {err}")

    def details(self, network: str, subnet: str, resource_group: str) -> Union[dict, None]:
        try:
            info = self.network_client.subnets.get(resource_group, network, subnet)
            subnet_block = {'cidr': info.address_prefix,
                            'name': info.name,
                            'routes': info.route_table.routes if info.route_table else None,
                            'nsg': info.network_security_group.id.rsplit('/', 1)[-1] if info.network_security_group else None,
                            'id': info.id}

            return subnet_block
        except ResourceNotFoundError:
            return None
        except Exception as err:
            raise AzureDriverError(f"error getting subnet: {err}")


class SecurityGroup(CloudBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger = logging.getLogger(self.__class__.__name__)

    def list(self, resource_group: Union[str, None] = None, filter_keys_exist: Union[List[str], None] = None) -> List[dict]:
        if not resource_group:
            resource_group = self.azure_resource_group
        nsg_list = []

        try:
            result = self.network_client.network_security_groups.list(resource_group)
        except Exception as err:
            raise AzureDriverError(f"error getting vnet: {err}")

        for group in list(result):
            if group.location != self.azure_location:
                continue
            nsg_block = {'location': group.location,
                         'name': group.name,
                         'rules': [r.__dict__ for r in group.security_rules] if group.security_rules else [],
                         'subnets': [s.__dict__ for s in group.subnets] if group.subnets else [],
                         'id': group.id}
            nsg_block.update(self.process_tags(group.tags))
            if filter_keys_exist:
                if not all(key in nsg_block for key in filter_keys_exist):
                    continue
            nsg_list.append(nsg_block)

        if len(nsg_list) == 0:
            raise EmptyResultSet(f"no suitable network security group in group {resource_group}")

        return nsg_list

    def create(self, name: str, resource_group: str):
        parameters = {
            'location': self.azure_location
        }
        try:
            request = self.network_client.network_security_groups.begin_create_or_update(resource_group, name, parameters)
            request.wait()
            return request.result()
        except Exception as err:
            raise AzureDriverError(f"error creating network security group: {err}")

    def add_rule(self,
                 name: str,
                 nsg_name: str,
                 ports: list,
                 priority: int,
                 resource_group: str,
                 protocol: str = "Tcp",
                 source: Union[list, None] = None) -> None:
        nsg_info = self.details(nsg_name, resource_group)
        if not nsg_info:
            raise AzureDriverError(f"can not find NSG {nsg_name} in {resource_group}")
        if source:
            default_source = None
        else:
            default_source = "*"
        protocol = protocol.lower().capitalize()
        if priority == 0:
            count = len(nsg_info.get('rules', []))
            priority = (count + 101)
        parameters = {
            "description": "Cloud Formation Managed",
            "access": "Allow",
            "destination_address_prefix": "*",
            "destination_port_ranges": ports,
            "direction": "Inbound",
            "priority": priority,
            "protocol": protocol,
            "source_address_prefix": default_source,
            "source_address_prefixes": source,
            "source_port_range": "*",
        }
        try:
            request = self.network_client.security_rules.begin_create_or_update(resource_group, nsg_name, name, parameters)
            request.wait()
            return request.result()
        except Exception as err:
            raise AzureDriverError(f"error creating network security group rule: {err}")

    def delete(self, name: str, resource_group: str) -> None:
        try:
            request = self.network_client.network_security_groups.begin_delete(resource_group, name)
            request.wait()
        except ResourceNotFoundError:
            return None
        except Exception as err:
            raise AzureDriverError(f"error getting network security group: {err}")

    def search_rules(self, name: str, resource_group: str, pattern: str) -> List[dict]:
        rules = []
        nsg_info = self.details(name, resource_group)
        if not nsg_info:
            raise AzureDriverError(f"can not find NSG {name} in {resource_group}")
        for rule in nsg_info.get('rules', []):
            if re.search(pattern, rule['name']):
                rules.append(rule)
        return rules

    def details(self, name: str, resource_group: str) -> Union[dict, None]:
        try:
            info = self.network_client.network_security_groups.get(resource_group, name)
            nsg_block = {'location': info.location,
                         'name': info.name,
                         'rules': [r.__dict__ for r in info.security_rules] if info.security_rules else [],
                         'subnets': [s.__dict__ for s in info.subnets] if info.subnets else [],
                         'id': info.id}
            nsg_block.update(self.process_tags(info.tags))
            return nsg_block
        except ResourceNotFoundError:
            return None
        except Exception as err:
            raise AzureDriverError(f"error getting network security group: {err}")



##
##

import logging
import concurrent.futures
import couchformation.executor.worker as worker
from couchformation.exception import NonFatalLogError

logger = logging.getLogger('couchformation.executor.dispatch')
logger.addHandler(logging.NullHandler())


class TaskError(NonFatalLogError):
    pass


class JobDispatch(object):

    def __init__(self):
        self.executor = concurrent.futures.ThreadPoolExecutor()
        self.tasks = set()

    def dispatch(self, *args, **kwargs):
        self.tasks.add(self.executor.submit(worker.main, *args, **kwargs))

    @staticmethod
    def foreground(*args, **kwargs):
        return worker.main(*args, **kwargs)

    @staticmethod
    def get_class(*args, **kwargs):
        return worker.get_class(*args, **kwargs)

    @staticmethod
    def run_method(*args, **kwargs):
        return worker.run_method(*args, **kwargs)

    def join(self):
        while self.tasks:
            done, self.tasks = concurrent.futures.wait(self.tasks, return_when=concurrent.futures.ALL_COMPLETED)
            for task in done:
                try:
                    res = task.result()
                    logger.debug(f"task result: {res}")
                    yield res
                except Exception as err:
                    raise TaskError(f"task exception: {err}")
